from __future__ import annotations

import os
import sys
from pathlib import Path
from typing import Dict

import requests
from rich.progress import wrap_file

from ..errors import ErrorPrinter


def __get_uploaded_bytes(upload_url: str) -> int:
    headers = {"Content-Range": "bytes */*"}
    response = requests.put(upload_url, headers=headers, timeout=60)
    if response.status_code != 308:
        raise ValueError(f"Failed to retrieve upload status: {response.status_code} {response.text}")

    # Parse the Range header to get the last byte uploaded
    range_header = response.headers.get("Range")
    if range_header:
        last_byte = int(range_header.split("-")[1])
        return last_byte + 1
    return 0


def with_resumable_upload_signed_url(signed_url: str, source_file_name: str, content_type: str) -> None:
    """
    Upload file to file storage with signed url and resumable upload.
    Resumable upload will only work with the same URL and not if a new signed URL is requested with the
    same object id.
    :param content_type:
    :param signed_url:
    :param source_file_name:
    :return:
    """

    file_size = os.path.getsize(source_file_name)
    headers = {"x-goog-resumable": "start", "content-type": content_type}
    response = requests.post(signed_url, headers=headers, timeout=60)
    if response.status_code not in (200, 201, 308):
        ErrorPrinter.print_generic_error(f"Failed to upload file: {response.status_code} - {response.text}")
        sys.exit(1)

    upload_url = response.headers["Location"]

    # Check how many bytes have already been uploaded
    uploaded_bytes = __get_uploaded_bytes(upload_url)

    # Upload the remaining file in chunks
    # Not sure what a good chunk size is or if we even should have resumable uploads here, probably not..
    chunk_size = 256 * 1024 * 10
    # Upload the file in chunks
    with open(source_file_name, "rb") as f:
        with wrap_file(f, os.stat(source_file_name).st_size, description=f"Uploading {source_file_name}...") as file:
            file.seek(uploaded_bytes)  # Seek to the position of the last uploaded byte
            for chunk_start in range(uploaded_bytes, file_size, chunk_size):
                chunk_end = min(chunk_start + chunk_size, file_size) - 1
                chunk = file.read(chunk_end - chunk_start + 1)
                headers = {"Content-Range": f"bytes {chunk_start}-{chunk_end}/{file_size}"}
                response = requests.put(upload_url, headers=headers, data=chunk, timeout=60)
                if response.status_code not in (200, 201, 308):
                    ErrorPrinter.print_generic_error(f"Failed to upload file: {response.status_code} - {response.text}")
                    sys.exit(1)

    print(f"File {source_file_name} uploaded successfully.")


def upload_signed_url(signed_url: str, source_file_name: Path, headers: Dict[str, str]) -> None:
    """
    Upload file to file storage with signed url and resumable upload.
    Resumable upload will only work with the same URL and not if a new signed URL is requested with the
    same object id.
    :param headers:
    :param signed_url:
    :param source_file_name:
    :return:
    """
    with open(source_file_name, "rb") as file:
        with wrap_file(file, os.stat(source_file_name).st_size, description=f"Uploading {source_file_name}...") as f:
            response = requests.put(signed_url, headers=headers, timeout=60, data=f)
            if response.status_code not in (200, 201, 308):
                ErrorPrinter.print_generic_error(f"Failed to upload file: {response.status_code} - {response.text}")
                sys.exit(1)

    print(f"File {source_file_name} uploaded successfully.")



from __future__ import annotations

import json
from pathlib import Path

from cli.cloud.resumable_upload import upload_signed_url
from cli.cloud.uri import URI
from cli.utils.rest_helper import RestHelper as Rest

_RCS_STORAGE_PATH = "/api/project/{project}/files/storage{path}"


def copy(project: str, source: URI | Path, dest: URI | Path, overwrite: bool = False) -> None:
    if isinstance(source, Path) and isinstance(dest, Path):
        raise ValueError("Either source or destination must be an rcs:// uri")

    if isinstance(source, URI) and isinstance(dest, URI):
        raise ValueError("Either source or destination must be a local path")

    if isinstance(source, URI) and isinstance(dest, Path):
        _download(source=source, dest=dest, project=project, overwrite=overwrite)

    elif isinstance(source, Path) and isinstance(dest, URI):
        _upload(source=source, dest=dest, project=project, overwrite=overwrite)

    else:
        raise ValueError("invalid copy operation")


def _upload(source: Path, dest: URI, project: str, overwrite: bool = False) -> None:
    if not source.exists():
        raise FileNotFoundError(f"Source file does not exist: {source}")

    files_to_upload = _list_files_for_upload(source, dest)

    for file_path, target_uri in files_to_upload:
        _upload_single_file(file_path, target_uri, project, overwrite)


def _list_files_for_upload(source: Path, dest: URI) -> list[tuple[Path, URI]]:
    upload_pairs = []

    if source.is_dir():
        for file_path in source.rglob("*"):
            if file_path.is_file():
                relative_path = file_path.relative_to(source)
                target_uri = dest / relative_path
                upload_pairs.append((file_path, target_uri))
    else:
        target_uri = dest / source.name if dest.is_dir() else dest
        upload_pairs.append((source, target_uri))

    return upload_pairs


def _upload_single_file(source: Path, target_uri: URI, project: str, overwrite: bool = False) -> None:
    target = _RCS_STORAGE_PATH.format(project=project, path=target_uri.path)
    upload_options = {"overwrite": "always" if overwrite else "never"}
    res = Rest.handle_post(target, return_response=True, body=json.dumps(upload_options))

    json_res = res.json()
    url = json_res["url"]
    headers = json_res["headers"]
    upload_signed_url(url, source, headers)

    print(f"Uploaded {source} to {target_uri.path}")


def _download(source: URI, dest: Path, project: str, overwrite: bool = False) -> None:
    if dest.is_dir():
        if not dest.exists():
            raise FileNotFoundError(f"Destination directory {dest} does not exist")
        # create a target file name if destination is a dir
        dest = dest / source.filename

    elif not dest.parent.is_dir() or not dest.parent.exists():
        raise FileNotFoundError(f"Destination directory {dest.parent} does not exist")

    if dest.exists() and not overwrite:
        raise FileExistsError(f"Destination file {dest} already exists")

    target = _RCS_STORAGE_PATH.format(project=project, path=source.path) + "?download=true"
    res = Rest.handle_get(target, return_response=True)

    Rest.download_file(save_file_name=dest.absolute(), url=res.text)



import json

import typer

from cli.typer import typer_utils
from cli.utils.rest_helper import RestHelper as Rest

app = typer_utils.create_typer()


@app.command(name="list", help="List your projects")
def list_projects(organisation: str = typer.Option(..., help="Organisation ID", envvar="REMOTIVE_CLOUD_ORGANISATION")) -> None:
    r = Rest.handle_get(url=f"/api/bu/{organisation}/me", return_response=True)
    if r is None:
        return

    if r.status_code == 200:
        # extract the project uid parts
        projects = r.json()["projects"]
        projects = map(lambda p: p["uid"], projects)
        print(json.dumps(list(projects)))
    else:
        print(r.status_code)
        print(r.text)


@app.command(name="create")
def create_project(
    project_uid: str = typer.Argument(help="Project UID"),
    organisation: str = typer.Option(..., help="Organisation ID", envvar="REMOTIVE_CLOUD_ORGANISATION"),
    project_display_name: str = typer.Option(default="", help="Project display name"),
) -> None:
    create_project_req = {
        "uid": project_uid,
        "displayName": project_display_name if project_display_name != "" else project_uid,
        "description": "",
    }

    Rest.handle_post(url=f"/api/bu/{organisation}/project", body=json.dumps(create_project_req))


@app.command(name="delete")
def delete(project: str = typer.Argument(help="Project ID", envvar="REMOTIVE_CLOUD_PROJECT")) -> None:
    Rest.handle_delete(url=f"/api/project/{project}")



from __future__ import annotations

import base64
import hashlib
import json
import os
import secrets
import sys
import time
import webbrowser
from http.server import BaseHTTPRequestHandler, HTTPServer
from threading import Thread
from typing import Any, Optional, Tuple
from urllib.parse import parse_qs, urlparse

import typer
from rich.console import Console
from typing_extensions import override

from cli.cloud.auth_tokens import list_and_select_personal_token
from cli.errors import ErrorPrinter
from cli.settings import TokenNotFoundError, settings
from cli.utils.rest_helper import RestHelper as Rest

httpd: HTTPServer

console = Console()


def generate_pkce_pair() -> Tuple[str, str]:
    """
    PKCE is used for all cli login flows, both headless and browser.
    """
    code_verifier_ = secrets.token_urlsafe(64)  # High-entropy string
    code_challenge_ = base64.urlsafe_b64encode(hashlib.sha256(code_verifier_.encode("ascii")).digest()).rstrip(b"=").decode("ascii")
    return code_verifier_, code_challenge_


code_verifier, code_challenge = generate_pkce_pair()
state = secrets.token_urlsafe(16)

short_lived_token = None


class S(BaseHTTPRequestHandler):
    def _set_response(self) -> None:
        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.end_headers()

    @override
    def log_message(self, format: Any, *args: Any) -> None:
        return

    # Please do not change this into lowercase!
    @override
    def do_GET(self) -> None:  # type: ignore # noqa: PLR0912
        self._set_response()

        parsed_url = urlparse(self.path)

        # Get query parameters as a dict
        query_params = parse_qs(parsed_url.query)

        # Example: Get the value of the "error" parameter if it exists
        error_value = query_params.get("error", [None])[0]
        path = self.path
        auth_code = path[1:]  # Remotive /
        time.sleep(1)
        httpd.server_close()

        killerthread = Thread(target=httpd.shutdown)
        killerthread.start()
        if error_value is None:
            res = Rest.handle_get(
                f"/api/open/token?code={auth_code}&code_verifier={code_verifier}",
                return_response=True,
                skip_access_token=True,
                allow_status_codes=[401, 400],
            )
            if res.status_code != 200:
                ErrorPrinter.print_generic_error(
                    "Failed to fetch token. Please try again, if the problem persists please reach out to support@remotivelabs.com"
                )
                self.wfile.write(
                    "Failed to fetch token. Please try again, if the problem persists please reach out to support@remotivelabs.com".encode(
                        "utf-8"
                    )
                )
                sys.exit(1)
            self.wfile.write("Successfully setup CLI, return to your terminal to continue".encode("utf-8"))
            access_token = res.json()["access_token"]
            # token = tf.TokenFile(
            #    name="CLI_login_token",
            #    token=access_token,
            #    created=str(datetime.datetime.now().isoformat()),
            #    expires="unknown",
            # )

            global short_lived_token  # noqa: PLW0603
            short_lived_token = access_token

            # settings.add_and_activate_short_lived_cli_token(tf.dumps(token))
            # print("Successfully logged on, you are ready to go with cli")
        else:
            if error_value == "no_consent":
                self.wfile.write(
                    """
                Authorization was cancelled.<br/>
                To use RemotiveCLI, you need to grant access to your RemotiveCloud account.
                <br/><br/>
                Run `remotive cloud auth login` to try again.
                """.encode("utf-8")
                )
                ErrorPrinter.print_generic_error("You did not grant access to RemotiveCloud, login aborted")
            else:
                self.wfile.write(f"Unknown error {error_value}, please contact support@remotivelabs.com".encode("utf-8"))
                ErrorPrinter.print_generic_error(f"Unexpected error {error_value}, please contact support@remotivelabs.com")
            sys.exit(1)


def prepare_local_webserver(server_class: type = HTTPServer, handler_class: type = S, port: Optional[int] = None) -> None:
    if port is None:
        env_val = os.getenv("REMOTIVE_LOGIN_CALLBACK_PORT" or "")
        if env_val and env_val.isdigit():
            port = int(env_val)
        else:
            port = 0

    server_address = ("", port)
    global httpd  # noqa: PLW0603
    httpd = server_class(server_address, handler_class)


def create_personal_token() -> None:
    response = Rest.handle_post(
        url="/api/me/keys",
        return_response=True,
        body=json.dumps({"alias": "roine"}),
        access_token=short_lived_token,
    )
    token = response.json()
    email = token["account"]["email"]
    existing_file = settings.get_token_file_by_email(email=email)
    if existing_file is not None:
        # ErrorPrinter.print_hint(f"Revoking and deleting existing credentials [remove_me]{existing_file.name}")
        res = Rest.handle_patch(
            f"/api/me/keys/{existing_file.name}/revoke",
            quiet=True,
            access_token=short_lived_token,
            allow_status_codes=[400, 404],
        )
        if res is not None and res.status_code == 200:
            Rest.handle_delete(
                f"/api/me/keys/{existing_file.name}",
                quiet=True,
                access_token=short_lived_token,
            )
        settings.remove_token_file(existing_file.name)

    settings.add_personal_token(response.text, activate=True)

    print("Successfully logged on")


def login(headless: bool = False) -> bool:  # noqa: C901, PLR0912, PLR0915
    """
    Initiate login
    """

    #
    # Check login.md flowchart for better understanding
    #
    # 1. Check for active token valid and working credentials
    #
    try:
        activate_token = settings.get_active_token_file()

        if not activate_token.is_expired():
            if Rest.has_access("/api/whoami"):
                console.print(f"You are already signed in with valid credentials that expires in {activate_token.expires_in_days()} days")
                files = settings.list_personal_token_files()
                if len(files) > 0:
                    print("")
                    console.print("You have available credentials on disk, [bold]choose one or press q to login again[/bold]")
                    token_selected = list_and_select_personal_token(skip_prompt=False)
                    if token_selected:
                        return True
                # list_and_select_personal_token(skip_prompt=True)
                # print("")
                # typer.confirm("Are you sure you want to login again?", abort=True)
                # If we are here, user still wants to login
            else:
                settings.clear_active_token()
                raise TokenNotFoundError()
        else:
            # TODO - Cleanup token since expired
            pass

    except TokenNotFoundError:
        #
        # 2. If no token was found, let user choose an existing if exists
        #
        files = settings.list_personal_token_files()
        if len(files) > 0:
            print("")
            token_selected = list_and_select_personal_token(
                skip_prompt=False,
                info_message="You have available credentials on disk, choose one or press q to login again",
            )
            if token_selected:
                return True

    prepare_local_webserver()

    def force_use_webserver_callback() -> bool:
        env_val = os.getenv("REMOTIVE_LOGIN_FORCE_CALLBACK" or "no")
        if env_val and env_val == "yes":
            return True
        return False

    def login_with_callback_but_copy_url() -> None:
        """
        This will print a url the will trigger a callback later so the webserver must be up and running.
        """
        print("Copy the following link in a browser to login to cloud, and complete the sign-in prompts:")
        print("")

        url = (
            f"{Rest.get_base_frontend_url()}/login"
            f"?state={state}"
            f"&cli_version={Rest.get_cli_version()}"
            f"&response_type=code"
            f"&code_challenge={code_challenge}"
            f"&redirect_uri=http://localhost:{httpd.server_address[1]}"
        )
        console.print(url, style="bold")
        httpd.serve_forever()

    def login_headless() -> None:
        """
        Full headless, opens a browser and expects a auth code to be entered and exchanged for the token
        """
        print("Copy the following link in a browser to login to cloud, and complete the sign-in prompts:")
        print("")

        url = (
            f"{Rest.get_base_frontend_url()}/login"
            f"?state={state}"
            f"&cli_version={Rest.get_cli_version()}"
            f"&response_type=code"
            f"&code_challenge={code_challenge}"
        )
        console.print(url, style="bold")

        code = typer.prompt(
            "Once finished, enter the verification code provided in your browser",
            hide_input=False,
        )
        res = Rest.handle_get(
            f"/api/open/token?code={code}&code_verifier={code_verifier}",
            return_response=True,
            skip_access_token=True,
            allow_status_codes=[401],
        )
        if res.status_code == 401:
            ErrorPrinter.print_generic_error(
                "Failed to fetch token. Please try again, if the problem persists please reach out to support@remotivelabs.com"
            )
            sys.exit(1)
        access_token = res.json()["access_token"]
        # res = Rest.handle_get("/api/whoami", return_response=True, access_token=access_token)
        global short_lived_token  # noqa: PLW0603
        short_lived_token = access_token
        create_personal_token()
        # current_user = res.json()
        # token = tf.TokenFile(
        #    type="authorized_user",
        #    name="CLI_login_token",
        #    token=access_token,
        #    created=str(datetime.datetime.now().isoformat()),
        #    expires="unknown",
        #    account=TokenFileUser(email=current_user["email"], uid=current_user["uid"], project=None),
        # )
        # settings.add_and_activate_short_lived_cli_token(tf.dumps(token))
        # console.print("Successfully logged on, you are ready to go with cli", style="green bold")

    if headless and not force_use_webserver_callback():
        login_headless()
    elif headless and force_use_webserver_callback():
        login_with_callback_but_copy_url()
    else:
        could_open = webbrowser.open_new_tab(
            f"{Rest.get_base_frontend_url()}/login"
            f"?state={state}"
            f"&cli_version={Rest.get_cli_version()}"
            f"&response_type=code"
            f"&code_challenge={code_challenge}"
            f"&redirect_uri=http://localhost:{httpd.server_address[1]}"
        )

        if not could_open:
            print(
                "Could not open a browser on this machine, this is likely because you are in an environment where no browser is avaialble"
            )
            print("")
            if force_use_webserver_callback():
                login_with_callback_but_copy_url()
            else:
                login_headless()
        else:
            httpd.serve_forever()

        # Once we received our callback or code we are logged in and ready to go
        create_personal_token()

    return True



from __future__ import annotations

import json
import numbers
import os
import signal as os_signal
import sys
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Iterable, List, TypedDict, Union

import grpc
import plotext as plt  # type: ignore
import typer
from rich import print as rich_rprint

from cli.errors import ErrorPrinter
from cli.typer import typer_utils

from .lib.broker import Broker, SubscribableSignal

app = typer_utils.create_typer(help=help)

DEFAULT_GRPC_URL = "http://localhost:50051"

# signal_values:list = list()


class Signals(TypedDict):
    name: str
    signals: List[Any]


signal_values: Dict[Any, Any] = {}


@app.command(name="list")
def list_signals(
    url: str = typer.Option(DEFAULT_GRPC_URL, help="Broker URL", envvar="REMOTIVE_BROKER_URL"),
    api_key: str = typer.Option(None, help="Cloud Broker API-KEY or access token", envvar="REMOTIVE_BROKER_API_KEY"),
    name_starts_with: Union[str, None] = typer.Option(None, help="Signal name prefix to include"),
    name_ends_with: Union[str, None] = typer.Option(None, help="Signal name suffix to include"),
) -> None:
    """
    List signal metadata on a broker

    Filter are inclusive so --name-starts-with and --name-ends-with will include name that matches both
    """
    try:
        broker = Broker(url, api_key)
        available_signals = broker.list_signal_names(prefix=name_starts_with, suffix=name_ends_with)
        print(json.dumps(available_signals))
    except grpc.RpcError as rpc_error:
        ErrorPrinter.print_grpc_error(rpc_error)


def read_scripted_code_file(file_path: Path) -> bytes:
    try:
        print(file_path)
        with open(file_path, "rb") as file:
            return file.read()
    except FileNotFoundError:
        print("File not found. Please check your file path.")
        sys.exit(1)


@app.command()
def subscribe(  # noqa: C901, PLR0913, PLR0915
    url: str = typer.Option(DEFAULT_GRPC_URL, help="Broker URL", envvar="REMOTIVE_BROKER_URL"),
    api_key: str = typer.Option(None, help="Cloud Broker API-KEY or access token", envvar="REMOTIVE_BROKER_API_KEY"),
    signal: List[str] = typer.Option([], help="Signal names to subscribe to, mandatory when not using script"),
    script: Path = typer.Option(
        None,
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
        help="Supply a path to Lua script that to use for signal transformation",
    ),
    on_change_only: bool = typer.Option(default=False, help="Only get signal if value is changed"),
    x_plot: bool = typer.Option(default=False, help="Experimental: Plot the signal in terminal. Note graphs are not aligned by time"),
    x_plot_size: int = typer.Option(default=100, help="Experimental: how many points show for each plot"),
    # samples: int = typer.Option(default=0, he)
) -> None:
    """
    Subscribe to a selection of signals

    Subscribe to two signals and have it printed to terminal
    ```
    remotive broker signals subscribe  --url http://localhost:50051 --signal can1:signal1 --signal can0:signal2
    ```

    Subscribe using a LUA script with signal transformations, read more about scripted signals at https://docs.remotivelabs.com/docs/remotive-broker
    ```
    remotive broker signals subscribe  --url http://localhost:50051 --script myvss_script.lua
    ```
    """

    if script is None:
        if len(signal) == 0:
            ErrorPrinter.print_generic_error("You must use --signal or use --script when subscribing")
            sys.exit(1)

    if script is not None:
        if len(signal) > 0:
            ErrorPrinter.print_generic_error("You must must not specify --signal when using --script")
            sys.exit(1)

    plt.title("Signals")

    def exit_on_ctrlc(_sig: Any, _frame: Any) -> None:
        os._exit(0)

    def on_frame_plot(x: Iterable[Any]) -> None:
        plt.clt()  # to clear the terminal
        plt.cld()  # to clear the data only
        frames = list(x)
        plt.clf()
        plt.subplots(len(list(filter(lambda n: n.startswith("ts_"), signal_values.keys()))))
        plt.theme("pro")

        for frame in frames:
            name = frame["name"]

            if not isinstance(frame["value"], numbers.Number):
                # Skip non numberic values
                # TODO - would exit and print info message if I knew how to
                continue

            y = [frame["value"]]
            t = [frame["timestamp_us"]]

            if name not in signal_values:
                signal_values[name] = [None] * x_plot_size
                signal_values[f"ts_{name}"] = [None] * x_plot_size
            signal_values[name] = signal_values[name] + y
            signal_values[f"ts_{name}"] = signal_values[f"ts_{name}"] + t

            if len(signal_values[name]) > x_plot_size:
                signal_values[name] = signal_values[name][len(signal_values[name]) - x_plot_size :]

            if len(signal_values[f"ts_{name}"]) > x_plot_size:
                signal_values[f"ts_{name}"] = signal_values[f"ts_{name}"][len(signal_values[f"ts_{name}"]) - x_plot_size :]

        cnt = 1
        for key, value in signal_values.items():
            if not key.startswith("ts_"):
                plt.subplot(cnt, 1).plot(signal_values[f"ts_{key}"], value, label=key, color=cnt)
                cnt = cnt + 1
        plt.sleep(0.001)  # to add
        plt.show()

    def on_frame_print(x: Iterable[Any]) -> None:
        rich_rprint(json.dumps(list(x)))

    os_signal.signal(os_signal.SIGINT, exit_on_ctrlc)

    if x_plot:
        on_frame_func = on_frame_plot
    else:
        on_frame_func = on_frame_print

    try:
        if script is not None:
            script_src = read_scripted_code_file(script)
            broker = Broker(url, api_key)
            broker.subscribe_on_script(script_src, on_frame_func, on_change_only)
        else:

            def to_subscribable_signal(sig: str):
                arr = sig.split(":")
                if len(arr) != 2:
                    ErrorPrinter.print_hint(f"--signal must have format namespace:signal ({sig})")
                    sys.exit(1)
                return SubscribableSignal(namespace=arr[0], name=arr[1])

            signals_to_subscribe_to = list(map(to_subscribable_signal, signal))

            broker = Broker(url, api_key)
            broker.long_name_subscribe(signals_to_subscribe_to, on_frame_func, on_change_only)
        print("Subscribing to signals, press Ctrl+C to exit")
    except grpc.RpcError as rpc_error:
        ErrorPrinter.print_grpc_error(rpc_error)


@app.command(help="List namespaces on broker")
def namespaces(
    url: str = typer.Option(DEFAULT_GRPC_URL, help="Broker URL", envvar="REMOTIVE_BROKER_URL"),
    api_key: str = typer.Option(None, help="Cloud Broker API-KEY or access token", envvar="REMOTIVE_BROKER_API_KEY"),
) -> None:
    try:
        broker = Broker(url, api_key)
        namespaces_json = broker.list_namespaces()
        print(json.dumps(namespaces_json))
    except grpc.RpcError as rpc_error:
        ErrorPrinter.print_grpc_error(rpc_error)


@app.command()
def frame_distribution(
    url: str = typer.Option(..., help="Broker URL", envvar="REMOTIVE_BROKER_URL"),
    api_key: str = typer.Option(None, help="Cloud Broker API-KEY or access token", envvar="REMOTIVE_BROKER_API_KEY"),
    namespace: str = typer.Option(..., help="Namespace"),
) -> None:
    """
    Use this command to get frames currently available on the specified namespace.
    """
    try:
        broker = Broker(url, api_key)

        def on_data(data: Dict[str, Any]) -> None:
            timestamp: str = datetime.now().strftime("%H:%M:%S")
            distribution = data["countsByFrameId"]
            if len(distribution) == 0:
                ErrorPrinter.print_hint(f"{timestamp} - No frames available")
            else:
                for d in distribution:
                    ErrorPrinter.print_generic_message(f"{timestamp}: {d}")

        broker.listen_on_frame_distribution(namespace, on_data)
    except grpc.RpcError as rpc_error:
        ErrorPrinter.print_grpc_error(rpc_error)



import sys

import typer
from typing_extensions import Annotated

from cli.cloud.storage.copy import copy
from cli.cloud.storage.uri_or_path import UriOrPath
from cli.cloud.storage.uri_or_path import uri as uri_parser
from cli.cloud.uri import URI, InvalidURIError, JoinURIError
from cli.errors import ErrorPrinter
from cli.typer import typer_utils
from cli.utils.rest_helper import RestHelper as Rest

HELP = """
Manage files ([yellow]Beta feature not available for all customers[/yellow])

Copy file from local to remote storage and vice versa, list and delete files.

"""

app = typer_utils.create_typer(rich_markup_mode="rich", help=HELP)


@app.command(name="ls")
def list_files(
    project: Annotated[str, typer.Option(help="Project ID", envvar="REMOTIVE_CLOUD_PROJECT", show_default=False)],
    uri: Annotated[URI, typer.Argument(help="Remote storage path", parser=URI)] = "rcs://",  # type: ignore
) -> None:
    """
    Listing remote files

    This will list files and directories in project top level directory
    remotive cloud storage ls rcs://

    This will list all files and directories matching the path
    remotive cloud storage ls rcs://fileOrDirectoryPrefix

    This will list all files and directories in the specified directory
    remotive cloud storage ls rcs://fileOrDirectory/
    """
    Rest.handle_get(f"/api/project/{project}/files/storage{uri.path}")


@app.command(name="rm")
def delete_file(
    project: Annotated[str, typer.Option(help="Project ID", envvar="REMOTIVE_CLOUD_PROJECT", show_default=False)],
    uri: Annotated[URI, typer.Argument(help="Remote storage path", parser=URI, show_default=False)],
) -> None:
    """
    [red]Deletes[/red] a file from remote storage, this cannot be undone :fire:

    [white]remotive cloud storage rm rcs://directory/filename[/white]
    """
    Rest.handle_delete(f"/api/project/{project}/files/storage{uri.path}")


@app.command(name="cp")
def copy_file(
    project: Annotated[str, typer.Option(help="Project ID", envvar="REMOTIVE_CLOUD_PROJECT", show_default=False)],
    source: Annotated[UriOrPath, typer.Argument(help="Remote or local path to source file", parser=uri_parser, show_default=False)],
    dest: Annotated[UriOrPath, typer.Argument(help="Remote or local path to destination file", parser=uri_parser, show_default=False)],
    overwrite: Annotated[bool, typer.Option(help="Overwrite existing file on RCS")] = False,
) -> None:
    """
    Copies a file to or from remote storage

    remotive cloud storage cp rcs://dir/filename .
    remotive cloud storage cp rcs://dir/filename filename

    remotive cloud storage cp filename rcs://dir/
    remotive cloud storage cp filename rcs://dir/filename
    """
    try:
        return copy(project=project, source=source.value, dest=dest.value, overwrite=overwrite)
    except (InvalidURIError, JoinURIError, ValueError, FileNotFoundError, FileExistsError) as e:
        ErrorPrinter.print_hint(str(e))
        sys.exit(1)



from __future__ import annotations

import json
from typing import List

import typer

from cli.typer import typer_utils
from cli.utils.rest_helper import RestHelper as Rest

from . import service_account_tokens

app = typer_utils.create_typer()


@app.command(name="list", help="List service-accounts")
def list_service_accounts(project: str = typer.Option(..., help="Project ID", envvar="REMOTIVE_CLOUD_PROJECT")) -> None:
    Rest.handle_get(f"/api/project/{project}/admin/accounts")


ROLES_DESCRIPTION = """
    [bold]Supported roles [/bold]
    project/admin          - Full project support, view, edit, delete and manage users
    project/user           - View, edit, upload, delete but no admin
    project/viewer         - View only
    project/storageCreator - Can upload to storage but not view, overwrite or delete
    org/topologyRunner     - Can start RemotiveTopology
"""


@app.command(
    name="create",
    help=f"""
    Create a new service account with one or more roles.

    [bold]Must only use lowercase letters, digits, or dashes, and must not contain -- or end with a dash.[/bold]

    remotive cloud service-accounts create --role project/user --role project/storageCreator
    {ROLES_DESCRIPTION}
    """,
)
def create_service_account(
    name: str,
    role: List[str] = typer.Option(..., help="Roles to apply"),
    project: str = typer.Option(..., help="Project ID", envvar="REMOTIVE_CLOUD_PROJECT"),
) -> None:
    data = {"name": name, "roles": role}
    Rest.handle_post(url=f"/api/project/{project}/admin/accounts", body=json.dumps(data))


@app.command(
    name="update",
    help=f"""
    Update an existing service account with one or more roles.

    remotive cloud service-accounts update --role project/user --role project/storageCreator
    {ROLES_DESCRIPTION}
    """,
)
def update_service_account(
    service_account: str = typer.Argument(..., help="Service account name"),
    role: List[str] = typer.Option(..., help="Roles to apply"),
    project: str = typer.Option(..., help="Project ID", envvar="REMOTIVE_CLOUD_PROJECT"),
) -> None:
    Rest.handle_put(url=f"/api/project/{project}/admin/accounts/{service_account}", body=json.dumps({"roles": role}))


@app.command(name="delete", help="Delete service account")
def delete_service_account(name: str, project: str = typer.Option(..., help="Project ID", envvar="REMOTIVE_CLOUD_PROJECT")) -> None:
    Rest.handle_delete(url=f"/api/project/{project}/admin/accounts/{name}")


app.add_typer(service_account_tokens.app, name="tokens", help="Manage project service account access tokens")



from __future__ import annotations

import dataclasses
import json
from dataclasses import dataclass
from typing import Dict, Optional

from dacite import from_dict


def loads(data: str) -> ConfigFile:
    d = json.loads(data)
    return from_dict(ConfigFile, d)


def dumps(config: ConfigFile) -> str:
    return json.dumps(dataclasses.asdict(config), default=str)


@dataclass
class Account:
    credentials_name: str
    default_organization: Optional[str] = None
    # Add project as well


@dataclass
class ConfigFile:
    version: str = "1.0"
    active: Optional[str] = None
    accounts: Dict[str, Account] = dataclasses.field(default_factory=dict)

    def get_active_default_organisation(self) -> Optional[str]:
        active_account = self.get_active()
        return active_account.default_organization if active_account is not None else None

    def get_active(self) -> Optional[Account]:
        if self.active is not None:
            account = self.accounts.get(self.active)
            if account is not None:
                return account
            raise KeyError(f"Activated account {self.active} is not a valid account")
        return None

    def activate(self, email: str) -> None:
        account = self.accounts.get(email)

        if account is not None:
            self.active = email
        else:
            raise KeyError(f"Account {email} does not exists")

    def get_account(self, email: str) -> Optional[Account]:
        if self.accounts:
            return self.accounts[email]
        return None

    def remove_account(self, email: str) -> None:
        if self.accounts:
            self.accounts.pop(email, None)

    def init_account(self, email: str, token_name: str) -> None:
        if self.accounts is None:
            self.accounts = {}

        account = self.accounts.get(email)
        if not account:
            account = Account(credentials_name=token_name)
        else:
            account.credentials_name = token_name
        self.accounts[email] = account

    def set_account_field(self, email: str, default_organization: Optional[str] = None) -> ConfigFile:
        if self.accounts is None:
            self.accounts = {}

        account = self.accounts.get(email)
        if not account:
            raise KeyError(f"Account with email {email} has not been initialized with token")

        # Update only fields explicitly passed
        if default_organization is not None:
            account.default_organization = default_organization

        return self



from __future__ import annotations

import datetime
import json
import os
import sys
import tempfile
from pathlib import Path
from typing import Any, List, Union

import grpc
import rich
import typer
from rich import print as rich_print
from rich.progress import Progress, SpinnerColumn, TextColumn

from cli.errors import ErrorPrinter
from cli.utils.rest_helper import RestHelper as Rest

from ..broker.lib.broker import Broker, SubscribableSignal
from ..typer import typer_utils

app = typer_utils.create_typer(
    help="""
Support for playback of a recording on a cloud broker, make sure to always mount a recording first
"""
)


@app.command()
def play(
    recording_session: str = typer.Argument(..., help="Recording session id", envvar="REMOTIVE_CLOUD_RECORDING_SESSION"),
    broker: str = typer.Option(None, help="Broker to use"),
    project: str = typer.Option(..., help="Project ID", envvar="REMOTIVE_CLOUD_PROJECT"),
    show_progress: bool = typer.Option(False, help="Show progress after started playing"),
    repeat: bool = typer.Option(False, help="Repeat recording - must keep command running in terminal"),
) -> None:
    """
    Start playing a recording.
    There is no problem invoking play multiple times since if it is already playing the command will be ignored.
    Use --repeat to have the recording replayed when it reaches the end.
    """

    _do_change_playback_mode("play", recording_session, broker, project, progress_on_play=show_progress, repeat=repeat)


@app.command()
def pause(
    recording_session: str = typer.Argument(..., help="Recording session id", envvar="REMOTIVE_CLOUD_RECORDING_SESSION"),
    broker: str = typer.Option(None, help="Broker to use"),
    project: str = typer.Option(..., help="Project ID", envvar="REMOTIVE_CLOUD_PROJECT"),
) -> None:
    """
    Pause a recording
    """
    _do_change_playback_mode("pause", recording_session, broker, project)


@app.command()
def progress(
    recording_session: str = typer.Argument(..., help="Recording session id", envvar="REMOTIVE_CLOUD_RECORDING_SESSION"),
    broker: str = typer.Option(None, help="Broker to use"),
    project: str = typer.Option(..., help="Project ID", envvar="REMOTIVE_CLOUD_PROJECT"),
) -> None:
    """
    Shows progress of the recording playing.
    Use --repeat to have the recording replayed when it reaches the end.
    """
    _do_change_playback_mode("status", recording_session, broker, project)


@app.command()
def seek(
    recording_session: str = typer.Argument(..., help="Recording session id", envvar="REMOTIVE_CLOUD_RECORDING_SESSION"),
    seconds: int = typer.Option(..., min=0, help="Target offset in seconds"),
    broker: str = typer.Option(None, help="Broker to use"),
    project: str = typer.Option(..., help="Project ID", envvar="REMOTIVE_CLOUD_PROJECT"),
) -> None:
    """
    Seek seconds into a recording
    """
    _do_change_playback_mode("seek", recording_session, broker, project, seconds)


@app.command()
def stop(
    recording_session: str = typer.Argument(..., help="Recording session id", envvar="REMOTIVE_CLOUD_RECORDING_SESSION"),
    broker: str = typer.Option(None, help="Broker to use"),
    project: str = typer.Option(..., help="Project ID", envvar="REMOTIVE_CLOUD_PROJECT"),
) -> None:
    """
    Stop playing
    """
    _do_change_playback_mode("stop", recording_session, broker, project)


# Copied from signals.py
def read_scripted_code_file(file_path: Path) -> bytes:
    # typer checks that the Path exists
    with open(file_path, "rb") as file:
        return file.read()


@app.command()
def subscribe(  # noqa: PLR0913
    recording_session: str = typer.Argument(..., help="Recording session id", envvar="REMOTIVE_CLOUD_RECORDING_SESSION"),
    broker: str = typer.Option(None, help="Broker to use"),
    signal: List[str] = typer.Option(None, help="Signal names to subscribe to, mandatory when not using script"),
    script: Path = typer.Option(
        None,
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
        help="Supply a path to Lua script that to use for signal transformation",
    ),
    on_change_only: bool = typer.Option(default=False, help="Only get signal if value is changed"),
    project: str = typer.Option(..., help="Project ID", envvar="REMOTIVE_CLOUD_PROJECT"),
) -> None:
    """
    Allows you to subscribe to signals based on a mounted recording without knowing the broker URL.
    This simplifies when playing recordings from the cloud.

    Terminal plotting is not yet supported here so we refer to remotive broker signals subscribe --x-plot for this.
    """
    if script is None:
        if len(signal) == 0:
            ErrorPrinter.print_generic_error("You must use include at least one signal and one namespace or use script when subscribing")
            sys.exit(1)
    if script is not None:
        if len(signal) > 0:
            ErrorPrinter.print_generic_error("You must must not specify --signal when using --script")
            sys.exit(1)

    broker_client = _get_broker_info(project, recording_session, broker, "subscribe")

    try:
        if script is not None:
            script_src = read_scripted_code_file(script)
            broker_client.subscribe_on_script(script_src, lambda sig: rich_print(json.dumps(list(sig))), on_change_only)
        else:

            def to_subscribable_signal(sig: str):
                arr = sig.split(":")
                if len(arr) != 2:
                    ErrorPrinter.print_hint(f"--signal must have format namespace:signal ({sig})")
                    sys.exit(1)
                return SubscribableSignal(namespace=arr[0], name=arr[1])

            signals_to_subscribe_to = list(map(to_subscribable_signal, signal))
            broker_client.long_name_subscribe(signals_to_subscribe_to, lambda sig: rich_print(json.dumps(list(sig))), on_change_only)
            print("Subscribing to signals, press Ctrl+C to exit")
    except grpc.RpcError as rpc_error:
        ErrorPrinter.print_grpc_error(rpc_error)

    except Exception as e:
        ErrorPrinter.print_generic_error(str(e))
        sys.exit(1)


def _do_change_playback_mode(  # noqa: C901, PLR0913, PLR0912
    mode: str,
    recording_session: str,
    brokerstr: str,
    project: str,
    seconds: int | None = None,
    progress_on_play: bool = False,
    repeat: bool = False,
) -> None:
    response = Rest.handle_get(f"/api/project/{project}/files/recording/{recording_session}", return_response=True)
    if response is None:
        return
    r = json.loads(response.text)
    enabled_recordings: List[Any] = [r for r in r["recordings"] if r["enabled"]]
    files = [{"recording": rec["fileName"], "namespace": rec["metadata"]["namespace"]} for rec in enabled_recordings]

    broker_name = brokerstr if brokerstr is not None else "personal"
    response = Rest.handle_get(f"/api/project/{project}/brokers/{broker_name}", return_response=True, allow_status_codes=[404])
    if response is None:
        return
    if response.status_code == 404:
        broker_arg = ""
        if brokerstr is not None:
            broker_arg = f" --broker {brokerstr} --ensure-broker-started"
        ErrorPrinter.print_generic_error("You need to mount the recording before you play")
        ErrorPrinter.print_hint(f"remotive cloud recordings mount {recording_session}{broker_arg} --project {project}")
        sys.exit(1)

    broker_info = json.loads(response.text)
    broker = Broker(broker_info["url"], None)

    _verify_recording_on_broker(broker, recording_session, mode, project)

    if mode == "pause":
        broker.pause_play(files, True)
    elif mode == "play":
        broker.play(files, True)
        if progress_on_play or repeat:
            _track_progress(broker, repeat, files)
    elif mode == "seek":
        if seconds is not None:
            broker.seek(files, int(seconds * 1000000), True)
        else:
            broker.seek(files, 0, True)
    elif mode == "stop":
        broker.seek(files, 0, True)
    elif mode == "status":
        _track_progress(broker, repeat, files)
    else:
        raise ValueError(f"Illegal command {mode}")


def _track_progress(broker: Broker, repeat: bool, files: List[Any]) -> None:
    p = Progress(SpinnerColumn(), TextColumn("[progress.description]{task.description}"), transient=True)
    t = p.add_task("label", total=1)
    if repeat:
        rich.print(":point_right: Keep this command running in terminal to keep the recording play with repeat")
    with p:

        def print_progress(offset: int, total: int, current_mode: str) -> None:
            p.update(
                t,
                description=f"{(datetime.timedelta(seconds=offset))} / {(datetime.timedelta(seconds=total))} ({current_mode})",
            )

        broker.listen_on_playback(repeat, files, print_progress)


def _verify_recording_on_broker(broker: Broker, recording_session: str, mode: str, project: str) -> None:
    try:
        # Here we try to verify that we are operating on a recording that is mounted on the
        # broker so we can verify this before we try playback and can also present some good
        # error messages
        tmp = os.path.join(tempfile.gettempdir(), os.urandom(24).hex())
        broker.download(".cloud.context", tmp, True)
        with open(tmp, "r", encoding="utf8") as f:
            json_context = json.loads(f.read())
            if json_context["recordingSessionId"] != recording_session:
                ErrorPrinter.print_generic_error(
                    f"The recording id mounted is '{json_context['recordingSessionId']}' "
                    f"which not the same as you are trying to {mode}, use cmd below to mount this recording"
                )
                ErrorPrinter.print_hint(f"remotive cloud recordings mount {recording_session} --project {project}")
                sys.exit(1)
    except grpc.RpcError as rpc_error:
        if rpc_error.code() == grpc.StatusCode.NOT_FOUND:
            ErrorPrinter.print_generic_error(f"You must use mount to prepare a recording before you can use {mode}")
            ErrorPrinter.print_hint(f"remotive cloud recordings mount {recording_session} --project {project}")
        else:
            ErrorPrinter.print_grpc_error(rpc_error)
        sys.exit(1)


def _get_broker_info(project: str, recording_session: str, broker: Union[str, None], mode: str) -> Broker:
    # Verify it exists
    Rest.handle_get(f"/api/project/{project}/files/recording/{recording_session}", return_response=True)

    broker_name = broker if broker is not None else "personal"
    response = Rest.handle_get(f"/api/project/{project}/brokers/{broker_name}", return_response=True, allow_status_codes=[404])
    if response is None:
        ErrorPrinter.print_generic_error(f"No response from: /api/project/{project}/brokers/{broker_name}")
        sys.exit(1)
    if response.status_code == 404:
        broker_arg = ""
        if broker is not None:
            broker_arg = f"--broker {broker} --ensure-broker-started"
        ErrorPrinter.print_generic_error("You need to mount the recording before you play")
        ErrorPrinter.print_hint(f"remotive cloud recordings mount {recording_session} {broker_arg} --project {project}")
        sys.exit(1)
    broker_info = json.loads(response.text)
    broker_client = Broker(broker_info["url"], None)
    _verify_recording_on_broker(broker_client, recording_session, mode, project)
    return broker_client



from __future__ import annotations

import sys

import typer

from cli.cloud.auth.login import login as do_login
from cli.errors import ErrorPrinter
from cli.settings import TokenNotFoundError, settings
from cli.typer import typer_utils
from cli.utils.rest_helper import RestHelper as Rest

from .. import auth_tokens
from ..organisations import do_select_default_org

HELP = """
Manage how you authenticate with our cloud platform
"""
app = typer_utils.create_typer(help=HELP)
# app.add_typer(auth_tokens.app, name="credentials", help="Manage account credentials")


@app.command(name="login")
def login(browser: bool = typer.Option(default=True, help="Does not automatically open browser, instead shows a link")) -> None:
    """
    Login to the cli using browser

    If not able to open a browser it will show fallback to headless login and show a link that
    users can copy into any browser when this is unsupported where running the cli - such as in docker,
    virtual machine or ssh sessions.

    This will be used as the current access token in all subsequent requests. This would
    be the same as activating a personal access key or service-account access key.
    """
    do_login(headless=not browser)
    if settings.get_cli_config().get_active_default_organisation() is None:
        set_default_organisation = typer.confirm(
            "You have not set a default organisation\nWould you like to choose one now?", abort=False, default=True
        )
        if set_default_organisation:
            do_select_default_org(get=False)


@app.command()
def whoami() -> None:
    """
    Validates authentication and fetches your account information
    """
    try:
        Rest.handle_get("/api/whoami")
    except TokenNotFoundError as e:
        ErrorPrinter.print_hint(str(e))
        sys.exit(1)


@app.command()
def print_access_token() -> None:
    """
    Print current active token
    """
    try:
        print(settings.get_active_token())
    except TokenNotFoundError as e:
        ErrorPrinter.print_hint(str(e))
        sys.exit(1)


def print_access_token_file() -> None:
    """
    Print current active token and its metadata
    """
    try:
        print(settings.get_active_token_file())
    except TokenNotFoundError as e:
        ErrorPrinter.print_hint(str(e))
        sys.exit(1)


# @app.command(help="Clears active credentials")
def logout() -> None:
    settings.clear_active_token()
    print("Access token removed")


app.command("activate")(auth_tokens.select_personal_token)
app.command("list")(auth_tokens.list_pats_files)



from cli.typer import typer_utils

from .can.can import app as can_app

HELP_TEXT = """
CLI tools unrelated to cloud or broker
"""

app = typer_utils.create_typer(help=HELP_TEXT)
app.add_typer(can_app, name="can", help="CAN tools")



from __future__ import annotations

import os
from importlib.metadata import version

import typer
from rich import print as rich_print
from rich.console import Console
from trogon import Trogon  # type: ignore
from typer.main import get_group

from cli.settings.migrate_all_token_files import migrate_any_legacy_tokens

from .broker.brokers import app as broker_app
from .cloud.cloud_cli import app as cloud_app
from .connect.connect import app as connect_app
from .settings import settings
from .tools.tools import app as tools_app
from .typer import typer_utils

err_console = Console(stderr=True)

if os.getenv("GRPC_VERBOSITY") is None:
    os.environ["GRPC_VERBOSITY"] = "NONE"

app = typer_utils.create_typer(
    rich_markup_mode="rich",
    help="""
Welcome to RemotiveLabs CLI - Simplify and automate tasks for cloud resources and brokers

For documentation - https://docs.remotivelabs.com
""",
)

# settings.set_default_config_as_env()


def version_callback(value: bool) -> None:
    if value:
        my_version = version("remotivelabs-cli")
        typer.echo(my_version)
        raise typer.Exit()


def test_callback(value: int) -> None:
    if value:
        rich_print(value)
        raise typer.Exit()


def _migrate_old_tokens() -> None:
    tokens = settings.list_personal_tokens()
    tokens.extend(settings.list_service_account_tokens())
    if migrate_any_legacy_tokens(tokens):
        err_console.print("Migrated old credentials and configuration files, you may need to login again or activate correct credentials")


def _set_default_org_as_env() -> None:
    """
    If not already set, take the default organisation from file and set as env
    This has to be done early before it is read
    """
    if "REMOTIVE_CLOUD_ORGANISATION" not in os.environ:
        org = settings.get_cli_config().get_active_default_organisation()
        if org is not None:
            os.environ["REMOTIVE_CLOUD_ORGANISATION"] = org


@app.callback()
def main(
    _the_version: bool = typer.Option(None, "--version", callback=version_callback, is_eager=False, help="Print current version"),
) -> None:
    _set_default_org_as_env()
    _migrate_old_tokens()
    # Do other global stuff, handle other global options here


@app.command()
def tui(ctx: typer.Context) -> None:
    """
    Explore remotive-cli and generate commands with this textual user interface application
    """

    Trogon(get_group(app), click_context=ctx).run()


app.add_typer(broker_app, name="broker", help="Manage a single broker - local or cloud")
app.add_typer(
    cloud_app,
    name="cloud",
    help="Manage resources in RemotiveCloud",
)
app.add_typer(connect_app, name="connect", help="Integrations with other systems")
app.add_typer(tools_app, name="tools")



import typer

from cli.cloud import auth, brokers, configs, organisations, projects, recordings, sample_recordings, service_accounts, storage
from cli.typer import typer_utils
from cli.utils.rest_helper import RestHelper

app = typer_utils.create_typer()


@app.command(help="List licenses for an organisation")
def licenses(
    organisation: str = typer.Option(..., help="Organisation ID", envvar="REMOTIVE_CLOUD_ORGANISATION"),
    filter_option: str = typer.Option("all", help="all, valid, expired"),
) -> None:
    RestHelper.handle_get(f"/api/bu/{organisation}/licenses", {"filter": filter_option})


app.add_typer(organisations.app, name="organisations", help="Manage organizations")
app.add_typer(projects.app, name="projects", help="Manage projects")
app.add_typer(auth.app, name="auth")
app.add_typer(brokers.app, name="brokers", help="Manage cloud broker lifecycle")
app.add_typer(recordings.app, name="recordings", help="Manage recordings")
app.add_typer(configs.app, name="signal-databases", help="Manage signal databases")
app.add_typer(storage.app, name="storage")
app.add_typer(service_accounts.app, name="service-accounts", help="Manage project service account keys")
app.add_typer(sample_recordings.app, name="samples", help="Manage sample recordings")

if __name__ == "__main__":
    app()



from __future__ import annotations

import os
import re
import shutil
import stat
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Tuple, Union

from rich.console import Console

from cli.errors import ErrorPrinter
from cli.settings import config_file
from cli.settings import token_file as tf
from cli.settings.config_file import ConfigFile
from cli.settings.token_file import TokenFile

err_console = Console(stderr=True)


CONFIG_DIR_PATH = Path.home() / ".config" / "remotive"
INCORRECT_CONFIG_DIR_PATH = Path.home() / ".config" / ".remotive"
DEPRECATED_CONFIG_DIR_PATH = Path.home() / ".remotive"

CLI_CONFIG_FILE_NAME = "config.json"
ACTIVE_TOKEN_FILE_NAME = "cloud.secret.token"
PERSONAL_TOKEN_FILE_PREFIX = "personal-token-"
SERVICE_ACCOUNT_TOKEN_FILE_PREFIX = "service-account-token-"


TokenFileMetadata = Tuple[TokenFile, Path]


class InvalidSettingsFilePathError(Exception):
    """Raised when trying to access an invalid settings file or file path"""


class NotFoundError(Exception):
    """Raised when a token cannot be found in settings"""


class TokenNotFoundError(Exception):
    """Raised when a token cannot be found in settings"""


@dataclass()
class CliConfigFile:
    default_organisation: Union[str, None]


class Settings:
    """
    Settings for the remotive CLI
    """

    config_dir: Path

    def __init__(self, config_dir: Path, deprecated_config_dirs: list[Path] | None = None) -> None:
        self.config_dir = config_dir
        self._active_secret_token_path = self.config_dir / ACTIVE_TOKEN_FILE_NAME
        self._cli_config = self.config_dir / CLI_CONFIG_FILE_NAME

        if self.config_dir.exists():
            return

        # create the config dir and try to migrate legacy config dirs if they exist
        self.config_dir.mkdir(parents=True, exist_ok=True)
        if deprecated_config_dirs:
            for deprecated_config_dir in deprecated_config_dirs:
                self._migrate_legacy_config_dir(deprecated_config_dir)

    # def _write_properties(self, filepath: Path, props: CliConfigFile) -> None:
    #    with open(filepath, "w", encoding="utf-8") as file:
    #        # keys = sorted(props.keys()) if sort_keys else props.keys()
    #        # for key in keys:
    #        file.write(f"default_organisation={props.default_organisation}\n")

    def _read_properties(self, filepath: Path) -> CliConfigFile:
        props = {}
        with open(filepath, "r", encoding="utf-8") as file:
            for line_num, line in enumerate(file, start=1):
                line_stripped = line.strip()
                if not line_stripped or line_stripped.startswith("#"):
                    continue
                if "=" not in line_stripped:
                    raise ValueError(f"Invalid line format at line {line_num}: {line}")
                key, value = line_stripped.split("=", 1)
                key, value = key.strip(), value.strip()
                if key in props:
                    raise ValueError(f"Duplicate key '{key}' found at line {line_num}")
                props[key] = value
        if "default_organisation" in props:
            return CliConfigFile(default_organisation=props["default_organisation"])
        return CliConfigFile(default_organisation=None)

    def set_default_organisation(self, organisation: str) -> None:
        cli_config = self.get_cli_config()

        try:
            token = settings.get_active_token_file()
            cli_config.set_account_field(token.account.email, organisation)
            self._write_config_file(cli_config)
        except TokenNotFoundError:
            ErrorPrinter.print_hint("You must have an account activated in order to set default organisation")
            sys.exit(1)

    # def set_default_config_as_env(self) -> None:
    #    config = self.get_cli_config()
    #    if config.default_organisation is not None:
    #        if "REMOTIVE_CLOUD_ORGANISATION" not in os.environ:
    #            os.environ["REMOTIVE_CLOUD_ORGANISATION"] = config.default_organisation

    # def get_active_cli_account(self) -> Optional[Account]:
    #    try:
    #        token = self.get_active_token_file()
    #        config = self.get_cli_config()
    #        if token.account.email in config.accounts:
    #            return config.accounts[token.account.email]
    #        return None
    #    except TokenNotFoundError:
    #        ErrorPrinter.print_hint("Cannot get active config without activating account credentials first")
    #        sys.exit(1)

    def get_cli_config(self) -> ConfigFile:
        try:
            return self._read_config_file()
        except TokenNotFoundError:
            return ConfigFile()
            # self._write_properties(CONFIG_DIR_PATH / CLI_CONFIG_FILE_NAME, [])

    def get_active_token(self) -> str:
        """
        Get the current active token secret
        """
        token_file = self.get_active_token_file()
        return token_file.token

    def get_active_token_file(self) -> TokenFile:
        """
        Get the current active token file
        """

        active_account = self.get_cli_config().get_active()
        if active_account is not None:
            token_name = active_account.credentials_name
            return self.get_token_file(token_name)
        raise TokenNotFoundError
        # if not self._active_secret_token_path.exists():
        #    raise TokenNotFoundError("no active token file found")
        # return self._read_token_file(self._active_secret_token_path)

    def activate_token(self, token: TokenFile) -> None:
        """
        Activate a token by name or path

        The token secret will be set as the current active secret.
        """
        # token_file = self.get_token_file(name)
        cli_config = self.get_cli_config()
        cli_config.activate(token.account.email)
        # if token_file.account.email not in cli_config.accounts:
        #    cli_config.set_account_field(token_file.account.email)
        self._write_config_file(cli_config)
        # self._write_token_file(self._active_secret_token_path, token_file)

    def clear_active_token(self) -> None:
        """
        Clear the current active token
        """
        config = self.get_cli_config()
        config.active = None
        self._write_config_file(config)

    # self._active_secret_token_path.unlink(missing_ok=True)

    def get_token_file_by_email(self, email: str) -> Optional[TokenFile]:
        tokens = [t for t in self.list_personal_tokens() if t.account is not None and t.account.email == email]
        if len(tokens) > 0:
            return tokens[0]
        tokens = [t for t in self.list_service_account_tokens() if t.account is not None and t.account.email == email]
        if len(tokens) > 0:
            return tokens[0]
        return None

    def get_token_file(self, name: str) -> TokenFile:
        """
        Get a token file by name or path
        """
        if Path(name).exists():
            return self._read_token_file(Path(name))
        if Path(CONFIG_DIR_PATH / name).exists():
            return self._read_token_file(Path(CONFIG_DIR_PATH / name))

        return self._get_token_by_name(name)[0]

    def remove_token_file(self, name: str) -> None:
        """
        Remove a token file by name or path

        TODO: what about manually downloaded tokens?
        """
        if Path(name).exists():
            if self.config_dir not in Path(name).parents:
                raise InvalidSettingsFilePathError(f"cannot remove a token file not located in settings dir {self.config_dir}")
            return Path(name).unlink()

        # TODO: what about the active token?
        path = self._get_token_by_name(name)[1]
        # print("Deleting", path)
        print(path)
        return path.unlink()

    # def add_and_activate_short_lived_cli_token(self, token: str) -> TokenFile:
    #    """
    #    Activates a short lived token
    #    """
    #    token_file = tf.loads(token)
    #    self._write_token_file(self._active_secret_token_path, token_file)
    #    return token_file

    def add_personal_token(
        self,
        token: str,
        activate: bool = False,
        overwrite_if_exists: bool = False,
    ) -> TokenFile:
        """
        Add a personal token
        """
        token_file = tf.loads(token)

        def email_to_safe_filename(email: str) -> str:
            # Replace any invalid character with an underscore
            return re.sub(r'[<>:"/\\|?*]', "_", email)

        # From now, user will never be None when adding a token so in this case token_file.user is never None
        email = email_to_safe_filename(token_file.account.email) if token_file.account is not None else "unknown"
        file = f"{PERSONAL_TOKEN_FILE_PREFIX}{token_file.name}-{email}.json"
        path = self.config_dir / file
        if path.exists() and not overwrite_if_exists:
            raise FileExistsError(f"Token file already exists: {path}")

        self._write_token_file(path, token_file)
        cli_config = self.get_cli_config()
        cli_config.init_account(email=token_file.account.email, token_name=token_file.name)
        self._write_config_file(cli_config)

        if activate:
            self.activate_token(token_file)

        return token_file

    def list_personal_tokens(self) -> list[TokenFile]:
        """
        List all personal tokens
        """
        return [f[0] for f in self._list_personal_tokens()]

    def list_personal_token_files(self) -> list[Path]:
        """
        List paths to all personal token files
        """
        return [f[1] for f in self._list_personal_tokens()]

    def add_service_account_token(self, token: str) -> TokenFile:
        token_file = tf.loads(token)

        def email_to_safe_filename(email: str) -> str:
            # Replace any invalid character with an underscore
            return re.sub(r'[<>:"/\\|?*]', "_", email)

            # From now, user will never be None when adding a token so in this case token_file.user is never None

        email = email_to_safe_filename(token_file.account.email) if token_file.account is not None else "unknown"
        file = f"{SERVICE_ACCOUNT_TOKEN_FILE_PREFIX}{email}-{token_file.name}.json"
        path = self.config_dir / file

        self._write_token_file(path, token_file)
        cli_config = self.get_cli_config()
        cli_config.init_account(email=token_file.account.email, token_name=token_file.name)
        self._write_config_file(cli_config)

        # if activate:
        #    self.activate_token(token_file.account.email)

        return token_file

        # token_file = tf.loads(token)
        # file = f"{SERVICE_ACCOUNT_TOKEN_FILE_PREFIX}{service_account}-{token_file.name}.json"
        # path = self.config_dir / file
        # if path.exists():
        #    raise FileExistsError(f"Token file already exists: {path}")

        # self._write_token_file(path, token_file)
        # return token_file

    def list_service_account_tokens(self) -> list[TokenFile]:
        """
        List all service account tokens
        """
        return [f[0] for f in self._list_service_account_tokens()]

    def list_service_account_token_files(self) -> list[Path]:
        """
        List paths to all service account token files
        """
        return [f[1] for f in self._list_service_account_tokens()]

    def _list_personal_tokens(self) -> list[TokenFileMetadata]:
        return self._list_token_files(prefix=PERSONAL_TOKEN_FILE_PREFIX)

    def _list_service_account_tokens(self) -> list[TokenFileMetadata]:
        return self._list_token_files(prefix=SERVICE_ACCOUNT_TOKEN_FILE_PREFIX)

    def _get_token_by_name(self, name: str) -> TokenFileMetadata:
        token_files = self._list_token_files()
        matches = [token_file for token_file in token_files if token_file[0].name == name]
        if len(matches) != 1:
            raise TokenNotFoundError(f"Ambiguous token file name {name}, found {len(matches)} files")
        return matches[0]

    def _list_token_files(self, prefix: str = "") -> list[TokenFileMetadata]:
        """list all tokens with the correct prefix in the config dir, but omit files that are not token files"""

        def is_valid_token_file(path: Path) -> bool:
            has_correct_prefix = path.is_file() and path.name.startswith(prefix)
            is_active_secret = path == self._active_secret_token_path
            is_cli_config = path == self._cli_config
            return has_correct_prefix and not is_active_secret and not is_cli_config

        paths = [path for path in self.config_dir.iterdir() if is_valid_token_file(path)]

        return [(self._read_token_file(token_file), token_file) for token_file in paths]

    def _read_token_file(self, path: Path) -> TokenFile:
        data = self._read_file(path)
        return tf.loads(data)

    def _read_config_file(self) -> ConfigFile:
        data = self._read_file(self.config_dir / CLI_CONFIG_FILE_NAME)
        return config_file.loads(data)

    def _read_file(self, path: Path) -> str:
        if not path.exists():
            raise TokenNotFoundError(f"File could not be found: {path}")
        return path.read_text(encoding="utf-8")

    def _write_token_file(self, path: Path, token: TokenFile) -> Path:
        data = tf.dumps(token)
        path = self._write_file(path, data)
        os.chmod(path, stat.S_IRUSR | stat.S_IWUSR)
        return path

    def _write_config_file(self, config: ConfigFile) -> Path:
        data = config_file.dumps(config)
        path = self._write_file(self._cli_config, data)
        os.chmod(path, stat.S_IRUSR | stat.S_IWUSR)
        return path

    def _write_file(self, path: Path, data: str) -> Path:
        if self.config_dir not in path.parents:
            raise InvalidSettingsFilePathError(f"file {path} not in settings dir {self.config_dir}")
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_text(data, encoding="utf8")
        return path

    def _migrate_legacy_config_dir(self, path: Path) -> None:
        if not path.exists():
            return

        sys.stderr.write(f"migrating deprecated config directory {path} to {self.config_dir}\n")
        shutil.copytree(str(path), str(self.config_dir), dirs_exist_ok=True)
        secret = path / ACTIVE_TOKEN_FILE_NAME
        if secret.exists():
            sys.stderr.write(f"Removing old activated token {secret}")
            secret.unlink(missing_ok=True)
            # value = secret.read_text(encoding="utf-8").strip()
            # The existing token file might either be a token file, or simply a string. We handle both cases...
            # try:
            #    token = tf.loads(value)
            # except JSONDecodeError:
            #    token = tf.TokenFile(
            #        version="1.0",
            #        type="service-account" if value.startswith("sa") else "authorized_user",
            #        name="MigratedActiveToken",
            #        token=value,
            #        created=str(datetime.datetime.now().isoformat()),
            #        expires="unknown",
            #        account=TokenFileAccount(email="unknown@remotivecloud.com"),
            #    )
            # self.add_and_activate_short_lived_cli_token(tf.dumps(token))
        shutil.rmtree(str(path))


def create_settings() -> Settings:
    """Create remotive CLI config directory and return its settings instance"""
    return Settings(
        CONFIG_DIR_PATH,
        deprecated_config_dirs=[DEPRECATED_CONFIG_DIR_PATH, INCORRECT_CONFIG_DIR_PATH],
    )


settings = create_settings()
"""
Global/module-level settings instance. Module-level variables are only loaded once, at import time.

TODO: Migrate away from singleton instance
"""



from __future__ import annotations

from cli.settings import settings
from cli.settings.core import ACTIVE_TOKEN_FILE_NAME
from cli.settings.migrate_token_file import InvalidTokenError, UnsupportedTokenVersionError, migrate_legacy_token
from cli.settings.token_file import TokenFile, dumps


def _migrate_legacy_tokens(tokens: list[TokenFile]) -> tuple[list[TokenFile], set[str]]:
    """
    Determine which tokens can be updated and which should be removed.

    Returns:
        tuple of (updated_tokens, invalid_tokens)
    """
    updated_tokens: list[TokenFile] = []
    invalid_tokens: set[str] = set()

    for token in tokens:
        try:
            migrated_token = migrate_legacy_token(token)
            if migrated_token.version != token.version:
                updated_tokens.append(migrated_token)
        except (InvalidTokenError, UnsupportedTokenVersionError):
            # Token not valid or unsupported version, mark for removal
            invalid_tokens.add(token.name)

    return updated_tokens, invalid_tokens


def _write_updated_tokens(updated_tokens: list[TokenFile]) -> None:
    for updated_token in updated_tokens:
        settings.remove_token_file(name=updated_token.name)
        if updated_token.type == "authorized_user":
            settings.add_personal_token(dumps(updated_token), overwrite_if_exists=True)
        elif updated_token.type == "service_account":
            settings.add_service_account_token(dumps(updated_token))
        else:
            raise ValueError(f"Unsupported token type: {updated_token.type}")


def _remove_invalid_tokens(invalid_tokens: set[str]) -> None:
    for token_name in invalid_tokens:
        settings.remove_token_file(name=token_name)


def _remove_old_secret_file() -> bool:
    old_activated_secret_file = settings.config_dir / ACTIVE_TOKEN_FILE_NAME
    old_secret_exists = old_activated_secret_file.exists()
    if old_secret_exists:
        old_activated_secret_file.unlink(missing_ok=True)
    return old_secret_exists


def migrate_any_legacy_tokens(tokens: list[TokenFile]) -> bool:
    """
    Migrate any legacy tokens to the latest TokenFile format.

    Returns True if any tokens were migrated, False otherwise.
    """
    # Get tokens to update/remove
    updated_tokens, invalid_tokens = _migrate_legacy_tokens(tokens)

    # Perform file operations
    _write_updated_tokens(updated_tokens)
    _remove_invalid_tokens(invalid_tokens)

    # Remove old secret file if exists
    old_secret_removed = _remove_old_secret_file()
    if old_secret_removed:
        return True  # We migrated at least one token

    # only return True if we migrated at least one token
    return len(updated_tokens) + len(invalid_tokens) > 0



from __future__ import annotations

from typing import List, Optional, Union

import typer
from rich.console import Console
from rich.table import Table

from cli.api.cloud import tokens
from cli.cloud.organisations import do_select_default_org
from cli.errors import ErrorPrinter
from cli.settings import TokenFile, TokenNotFoundError, settings
from cli.typer import typer_utils
from cli.utils.rest_helper import RestHelper as Rest

console = Console(stderr=False)
err_console = Console(stderr=True)

app = typer_utils.create_typer()


def _prompt_choice(  # noqa: C901, PLR0912
    choices: List[TokenFile],
    skip_prompt: bool = False,
    info_message: Optional[str] = None,
) -> Optional[Union[TokenFile, None]]:
    accounts = settings.get_cli_config().accounts

    try:
        active_account = settings.get_cli_config().get_active()
    except TokenNotFoundError:
        active_account = None

    table = Table("#", "Active", "Type", "Token", "Account", "Created", "Expires")

    included_tokens: list[TokenFile] = []
    excluded_tokens: list[TokenFile] = []

    for token in choices:
        account = accounts.get(token.account.email)
        if account and account.credentials_name and account.credentials_name in (token.name or ""):
            included_tokens.append(token)
        else:
            excluded_tokens.append(token)
    if len(excluded_tokens) > 0:
        err_console.print("The following credentials were not included in account config and cannot be activated")
        for token in excluded_tokens:
            err_console.print(f" * {token.name} - {token.account.email}")

    if len(included_tokens) == 0:
        return None

    if info_message is not None:
        console.print(info_message)

    included_tokens.sort(key=lambda token: token.created, reverse=True)

    for idx, choice in enumerate(included_tokens, start=1):
        table.add_row(
            f"[yellow]{idx}",
            ":white_check_mark:" if active_account is not None and active_account.credentials_name == choice.name else "",
            "user" if choice.type == "authorized_user" else "sa",
            choice.name,
            f"[bold]{choice.account.email if choice.account else 'unknown'}[/bold]",
            str(choice.created),
            str(choice.expires),
        )

    # console.print("It seems like you have access tokens from previous login, you can select one of these instead of logging in")
    console.print(table)

    if skip_prompt:
        return None

    typer.echo("")
    selection = typer.prompt(f"Enter the number(# 1-{len(included_tokens)}) of the token to select (or q to skip)")

    if selection == "q":
        return None
    try:
        index = int(selection) - 1
        if 0 <= index < len(included_tokens):
            return included_tokens[index]
        raise ValueError
    except ValueError:
        typer.echo("Invalid choice, please try again")
        return _prompt_choice(included_tokens, skip_prompt, info_message)


# @app.command(name="create")
def create(
    activate: bool = typer.Option(False, help="Activate the token for use after download"),
) -> None:
    """
    Create a new personal access token in [bold]cloud[/bold] and download locally
    """
    response = tokens.create()
    pat = settings.add_personal_token(response.text())
    print(f"Personal access token added: {pat.name}")

    if not activate:
        print(f"Use 'remotive cloud auth tokens activate {pat.name}' to use this access token from cli")
    else:
        settings.activate_token(pat)
        print("Token file activated and ready for use")
    print("\033[93m This file contains secrets and must be kept safe")


# @app.command(name="list", help="List personal credentials in [bold]cloud[/bold]")
def list_tokens() -> None:
    Rest.handle_get("/api/me/keys")


# @app.command(name="revoke")
def revoke(
    name: str = typer.Argument(help="Access token name"),
    delete: bool = typer.Option(True, help="Also delete token"),
) -> None:
    """
    Revoke personal credentials in cloud and removes the file from filesystem

    If cloud token is not found but token is found on file system it will delete it and
    vice versa.
    """
    _revoke_and_delete_personal_token(name, delete)


# @app.command(name="activate")
def activate(
    token_name: str = typer.Argument(..., help="Token path, filename or name to activate"),
) -> None:
    """
    Activate a credential file to be used for authentication using filename, path or name.

    This will be used as the current access token in all requests.
    """
    try:
        token_file = settings.get_token_file(token_name)
        settings.activate_token(token_file)
    except TokenNotFoundError:
        err_console.print(f":boom: [bold red] Error: [/bold red] Token with filename or name {token_name} could not be found")


@app.command("activate")
def select_personal_token(
    token_name: str = typer.Argument(None, help="Name, filename or path to a credentials file"),
) -> None:
    """
    Prompts user to select and activate one of the available credential files
    """
    if token_name is not None:
        token_selected = True
        try:
            token_file = settings.get_token_file(token_name)
            settings.activate_token(token_file)
        except TokenNotFoundError:
            token_selected = False
            err_console.print(f":boom: [bold red] Error: [/bold red] Token with filename or name {token_name} could not be found")
    else:
        token_files = settings.list_personal_tokens()
        token_files.extend(settings.list_service_account_tokens())
        if len(token_files) > 0:
            token_selected = list_and_select_personal_token(include_service_accounts=True)

            if token_selected:
                is_logged_in = Rest.has_access("/api/whoami")
                if not is_logged_in:
                    ErrorPrinter.print_generic_error("Could not access RemotiveCloud with selected token")
                else:
                    console.print("[green]Success![/green] Access to RemotiveCloud granted")
        else:
            ErrorPrinter.print_hint("No credentials available, login to activate credentials")

    if settings.get_cli_config().get_active_default_organisation() is None:
        set_default_organisation = typer.confirm(
            "You have not set a default organisation\nWould you like to choose one now?",
            abort=False,
            default=True,
        )
        if set_default_organisation:
            do_select_default_org(get=False)


def list_and_select_personal_token(
    skip_prompt: bool = False,
    include_service_accounts: bool = False,
    info_message: Optional[str] = None,
) -> bool:
    personal_tokens = settings.list_personal_tokens()

    if include_service_accounts:
        sa_tokens = settings.list_service_account_tokens()
        personal_tokens.extend(sa_tokens)

    # merged = _merge_local_tokens_with_cloud(personal_tokens)

    selected_token = _prompt_choice(personal_tokens, skip_prompt=skip_prompt, info_message=info_message)
    if selected_token is not None:
        settings.activate_token(selected_token)
        return True
    return False


# @app.command("select-revoke")
def select_revoke_personal_token() -> None:
    """
    Prompts a user to select one of the credential files to revoke and delete
    """
    personal_tokens = settings.list_personal_tokens()
    sa_tokens = settings.list_service_account_tokens()
    personal_tokens.extend(sa_tokens)

    is_logged_in = Rest.has_access("/api/whoami")
    if not is_logged_in:
        ErrorPrinter.print_hint("You must be logged in")
        raise typer.Exit(0)

    # merged = _merge_local_tokens_with_cloud(personal_tokens)

    selected_token = _prompt_choice(personal_tokens)

    if selected_token is not None:
        _revoke_and_delete_personal_token(selected_token.name, True)
        # Rest.handle_patch(f"/api/me/keys/{selected_token.name}/revoke", quiet=True, access_token=selected_token.token)
        # Rest.handle_delete(f"/api/me/keys/{selected_token.name}", quiet=True, access_token=selected_token.token)
        # settings.remove_token_file(selected_token.name)
        # active_token = settings.get_active_token_file()
        # if active_token.name == selected_token.name:
        #    settings.clear_active_token()
        # select_revoke_personal_token()


# @app.command("test-all")
def test_all_personal_tokens() -> None:
    """
    Tests each credential file to see if it is valid
    """
    personal_tokens = settings.list_personal_tokens()
    personal_tokens.extend(settings.list_service_account_tokens())
    if len(personal_tokens) == 0:
        console.print("No personal tokens found on disk")
        return

    for token in personal_tokens:
        r = Rest.handle_get(
            "/api/whoami",
            allow_status_codes=[401],
            access_token=token.token,
            use_progress_indicator=True,
            return_response=True,
        )
        if r.status_code == 200:
            if token.account is not None:
                console.print(f"{token.account.email} ({token.name}) :white_check_mark:")
            else:
                console.print(f"{token.name} :white_check_mark:")
        elif token.account is not None:
            console.print(f"{token.account.email} ({token.name}) :x: Failed")
        else:
            console.print(f"{token.name} :x: Failed")


# @app.command(name="list-service-account-tokens-files")
def list_sats_files() -> None:
    """
    List service account access token files in remotivelabs config directory
    """
    service_account_files = settings.list_service_account_token_files()
    for file in service_account_files:
        print(file)


@app.command(name="list")
def list_pats_files() -> None:
    """
    Lists available credential files on filesystem
    """
    personal_files = settings.list_personal_token_files()
    service_account_files = settings.list_service_account_token_files()
    personal_files.extend(service_account_files)
    for file in personal_files:
        print(file)


def _revoke_and_delete_personal_token(name: str, delete: bool) -> None:
    token_file = None

    # First we try to find the file and make sure its not the currently active
    try:
        token_file = settings.get_token_file(name)
        active_token = settings.get_active_token_file()
        if token_file.name == active_token.name:
            ErrorPrinter.print_hint("You cannot revoke the current active token")
            return
    except TokenNotFoundError:
        pass

    # The lets try to revoke from cloud
    res_revoke = tokens.revoke(name)
    if delete:
        res_delete = tokens.delete(name)
        if res_delete.is_success:
            ErrorPrinter.print_generic_message("Token successfully revoked and deleted")
        else:
            ErrorPrinter.print_hint(f"Failed to revoke and delete token in cloud: {res_delete.status_code}")
    elif res_revoke.is_success:
        ErrorPrinter.print_generic_message("Token successfully revoked")
    else:
        ErrorPrinter.print_hint("Failed to revoke and delete token in cloud")

    # Finally try to remove the file if exists
    if token_file is not None:
        settings.remove_token_file(token_file.name)
        console.print("Successfully deleted token on filesystem")
    else:
        ErrorPrinter.print_hint("Token not found on filesystem")



from __future__ import annotations

import sys
from string import Template
from typing import List

import typer
from rich import print as rich_print

from cli.errors import ErrorPrinter
from cli.typer import typer_utils

app = typer_utils.create_typer(
    rich_markup_mode="rich",
    help="""
[Experimental] - Generate template lua script for input and output signals
""",
)


def write(signal_name: str, script: str) -> None:
    path = f"{signal_name}.lua"
    f = open(path, "w")
    f.write(script)
    f.close()
    print(f"Secret token written to {path}")


@app.command("new-script")
def new_script(
    input_signal: List[str] = typer.Option(..., help="Required input signal names"),
    output_signal: str = typer.Option(..., help="Name of output signal"),
    save: bool = typer.Option(False, help="Save file to disk - Default stored as __output_signal__.lua"),
) -> None:
    def to_subscribable_signal(sig: str) -> tuple[str, str]:
        arr = sig.split(":")
        if len(arr) != 2:
            ErrorPrinter.print_hint(f"--input-signal must have format namespace:signal ({sig})")
            sys.exit(1)
        return arr[0], arr[1]

    signals_to_subscribe_to = list(map(to_subscribable_signal, input_signal))

    def to_local_signal(sig_name: tuple[str, str]) -> str:
        t = Template(
            """
    {
        name = "$sig_name",
        namespace = "$namespace"
    }"""
        )
        return t.substitute(sig_name=sig_name[1], namespace=sig_name[0])

    local_signals = ",".join(list(map(to_local_signal, signals_to_subscribe_to)))

    def to_subscribe_pattern(sig_name: tuple[str, str]) -> str:
        t = Template(
            """
    if (signals["$sig_name"] ~= nil) then
        return return_value_or_bytes(signals["$sig_name"])
    end
    """
        )
        return t.substitute(sig_name=sig_name[1])

    subscribe_pattern = "".join(list(map(to_subscribe_pattern, signals_to_subscribe_to)))

    template = Template(
        """
--
-- Docs available at https://docs.remotivelabs.com/docs/remotive-broker/scripted_signals
--

local local_signals = {$local_signals
}

-- Required, declare which input is needed to operate this program.
function input_signals()
    return local_signals
end

-- Provided parameters are used for populating metadata when listing signals.
function output_signal()
    return "$output_signal"
end

-- Required, declare what frequency you like to get "timer" invoked. 0 means no calls to "timer".
function timer_frequency_hz()
    return 0
end

-- Invoked with the frequency returned by "timer_frequency_hz".
-- @param system_timestamp_us: system time stamp
function timer(system_timestamp_us)
    return return_value_or_bytes("your value")
end

-- Invoked when ANY signal declared in "local_signals" arrive
-- @param signals_timestamp_us: signal time stamp
-- @param system_timestamp_us
-- @param signals: array of signals containing all or a subset of signals declared in "local_signals". Make sure to nil check before use.
function signals(signals, namespace, signals_timestamp_us, system_timestamp_us)
    -- TODO - replace this code with what you want todo

    $subscribe_pattern
    return return_nothing()
end

-- helper return function, make sure to use return_value_or_bytes or return_nothing.
function return_value_or_bytes(value_or_bytes)
    return value_or_bytes
end

-- helper return function, make sure to use return_value_or_bytes or return_nothing.
function return_nothing()
    return
end

"""
    )

    script = template.substitute(
        local_signals=local_signals,
        # signle_input_signal=signals_to_subscribe_to[0][1],  # Take one signal and use signal name in tuple
        subscribe_pattern=subscribe_pattern,
        output_signal=output_signal,
    )

    if save:
        write(output_signal, script)
    else:
        rich_print(script)



from __future__ import annotations

import json
from dataclasses import dataclass
from typing import List, Optional

import typer
from rich.console import Console
from rich.table import Table

from cli.settings import settings
from cli.typer import typer_utils
from cli.utils.rest_helper import RestHelper

console = Console(stderr=False)
app = typer_utils.create_typer()


@dataclass
class Organisation:
    display_name: str
    uid: str


def _prompt_choice(choices: List[Organisation]) -> Optional[Organisation]:
    table = Table("#", "Name", "Uid", "Default")

    token = settings.get_active_token_file()
    config = settings.get_cli_config()

    current_default_org = config.accounts[token.account.email].default_organization if token is not None else None

    for idx, choice in enumerate(choices, start=1):
        table.add_row(
            f"[yellow]{idx}",
            f"[bold]{choice.display_name}[/bold]",
            choice.uid,
            ":thumbsup:" if current_default_org is not None and current_default_org == choice.uid else "",
        )
    console.print(table)

    typer.echo("")
    selection = typer.prompt(f"Enter the number(# 1-{len(choices)}) of the organisation to select (or q to quit)")

    if selection == "q":
        return None
    try:
        index = int(selection) - 1
        if 0 <= index < len(choices):
            return choices[index]
        raise ValueError
    except ValueError:
        typer.echo("Invalid choice, please try again")
        return _prompt_choice(choices)


@app.command("default")
def select_default_org(
    organisation_uid: str = typer.Argument(None, help="Organisation uid or empty to select one"),
    get: bool = typer.Option(False, help="Print current default organisation"),
) -> None:
    do_select_default_org(organisation_uid, get)


def do_select_default_org(organisation_uid: Optional[str] = None, get: bool = False) -> None:
    r"""
    Set default organisation for the currently activated user, empty to choose from available organizations or organization uid as argument

    remotive cloud organizations default my_org \[set specific org uid]
    remotive cloud organizations default \[select one from prompt]
    remotive cloud organizations default --get \[print current default]

    Note that service-accounts does Not have permission to list organizations and will get a 403 Forbidden response so you must
    select the organization uid as argument
    """
    if get:
        default_organisation = settings.get_cli_config().get_active_default_organisation()
        if default_organisation is not None:
            console.print(default_organisation)
        else:
            console.print("No default organisation set")
    elif organisation_uid is not None:
        settings.set_default_organisation(organisation_uid)
    else:
        r = RestHelper.handle_get("/api/bu", return_response=True)
        orgs = r.json()
        orgs = [Organisation(display_name=o["organisation"]["displayName"], uid=o["organisation"]["uid"]) for o in orgs]

        selected = _prompt_choice(orgs)

        if selected is not None:
            settings.get_cli_config()
            typer.echo(f"Default organisation: {selected.display_name} (uid: {selected.uid})")
            settings.set_default_organisation(selected.uid)


@app.command(name="list", help="List your available organizations")
def list_orgs() -> None:
    r = RestHelper.handle_get("/api/bu", return_response=True)
    orgs = [{"uid": org["organisation"]["uid"], "displayName": org["organisation"]["displayName"]} for org in r.json()]
    print(json.dumps(orgs))



from cli.settings.core import InvalidSettingsFilePathError, Settings, TokenNotFoundError, settings
from cli.settings.token_file import TokenFile

__all__ = ["settings", "TokenFile", "TokenNotFoundError", "InvalidSettingsFilePathError", "Settings"]



from __future__ import annotations

import binascii
import ntpath
import os
import posixpath
import queue
import signal as os_signal
import sys
import tempfile
import time
import zipfile
from dataclasses import dataclass
from threading import Thread
from typing import Any, Callable, Dict, Iterable, List, Sequence, Union

import grpc
import remotivelabs.broker.generated.sync.network_api_pb2 as network_api
import remotivelabs.broker.generated.sync.traffic_api_pb2 as traffic_api
import remotivelabs.broker.sync as br
import remotivelabs.broker.sync.helper as br_helper
import typer
from google.protobuf.json_format import MessageToDict
from rich.console import Console

from cli.errors import ErrorPrinter
from cli.settings import settings

err_console = Console(stderr=True)


@dataclass
class SubscribableSignal:
    name: str
    namespace: str


@dataclass
class LicenseInfo:
    valid: bool
    expires: str
    email: str
    machine_id: str


class Broker:
    def __init__(self, url: str, api_key: Union[str, None] = None) -> None:
        self.url = url
        self.api_key = api_key
        self.q: queue.Queue[Any] = queue.Queue()
        """Main function, checking arguments passed to script, setting up stubs, configuration and starting Threads."""
        # Setting up stubs and configuration

        if api_key is None or api_key == "":
            if url.startswith("https"):
                self.intercept_channel = br.create_channel(url, None, settings.get_active_token())
                # TODO - Temporary solution to print proper error message, remove ENV once api-key is gone
                os.environ["ACCESS_TOKEN"] = "true"
            else:
                os.environ["ACCESS_TOKEN"] = "false"
                self.intercept_channel = br.create_channel(url, None, None)
        else:
            err_console.print("Option --api-key will is deprecated and will be removed. Use access access tokens by logging in with cli.")
            os.environ["ACCESS_TOKEN"] = "false"
            self.intercept_channel = br.create_channel(url, api_key, None)

        self.network_stub = br.network_api_pb2_grpc.NetworkServiceStub(self.intercept_channel)
        self.system_stub = br.system_api_pb2_grpc.SystemServiceStub(self.intercept_channel)
        self.traffic_stub = br.traffic_api_pb2_grpc.TrafficServiceStub(self.intercept_channel)
        self.signal_creator = br.SignalCreator(self.system_stub)

    @staticmethod
    def __check_playbackmode_result(status: br.traffic_api_pb2.PlaybackInfos) -> None:
        err_cnt = 0
        for mode in status.playbackInfo:
            if mode.playbackMode.errorMessage:
                print(mode.playbackMode.errorMessage)
                err_cnt = err_cnt + 1
        if err_cnt > 0:
            raise typer.Exit(1)

    def seek(self, recording_and_namespace: List[Any], offset: int, silent: bool = True) -> br.traffic_api_pb2.PlaybackInfos:
        def to_playback(rec: Any) -> Dict[str, Any]:
            return {"namespace": rec["namespace"], "path": rec["recording"], "mode": br.traffic_api_pb2.Mode.SEEK, "offsettime": offset}

        playback_list = map(to_playback, recording_and_namespace)

        infos = br.traffic_api_pb2.PlaybackInfos(playbackInfo=list(map(self.__create_playback_config, playback_list)))
        status = self.traffic_stub.PlayTraffic(infos)
        if not silent:
            self.__check_playbackmode_result(status)
        return status

    def play(self, recording_and_namespace: List[Any], silent: bool = False) -> br.traffic_api_pb2.PlaybackInfos:
        def to_playback(rec: Any) -> Dict[str, Any]:
            return {
                "namespace": rec["namespace"],
                "path": rec["recording"],
                "mode": br.traffic_api_pb2.Mode.PLAY,
            }

        playback_list = map(to_playback, recording_and_namespace)

        status = self.traffic_stub.PlayTraffic(
            br.traffic_api_pb2.PlaybackInfos(playbackInfo=list(map(self.__create_playback_config, playback_list)))
        )

        if not silent:
            self.__check_playbackmode_result(status)
        return status

    def stop_play(self, recording_and_namespace: List[Any], silent: bool = False) -> br.traffic_api_pb2.PlaybackInfos:
        def to_playback(rec: Any) -> Dict[str, Any]:
            return {
                "namespace": rec["namespace"],
                "path": rec["recording"],
                "mode": br.traffic_api_pb2.Mode.STOP,
            }

        playback_list = map(to_playback, recording_and_namespace)

        status = self.traffic_stub.PlayTraffic(
            br.traffic_api_pb2.PlaybackInfos(playbackInfo=list(map(self.__create_playback_config, playback_list)))
        )
        if not silent:
            self.__check_playbackmode_result(status)
        return status

    def pause_play(self, recording_and_namespace: List[Any], silent: bool = False) -> br.traffic_api_pb2.PlaybackInfos:
        def to_playback(rec: Any) -> Dict[str, Any]:
            return {
                "namespace": rec["namespace"],
                "path": rec["recording"],
                "mode": br.traffic_api_pb2.Mode.PAUSE,
            }

        playback_list = map(to_playback, recording_and_namespace)

        status = self.traffic_stub.PlayTraffic(
            br.traffic_api_pb2.PlaybackInfos(playbackInfo=list(map(self.__create_playback_config, playback_list)))
        )
        if not silent:
            self.__check_playbackmode_result(status)
        return status

    def record_multiple(self, namespaces: List[str], path: str) -> br.traffic_api_pb2.PlaybackInfos:
        def to_playback(namespace: str) -> Dict[str, Any]:
            return {
                "namespace": namespace,
                "path": path + "_" + namespace,
                "mode": br.traffic_api_pb2.Mode.RECORD,
            }

        playback_list = list(map(to_playback, namespaces))

        status = self.traffic_stub.PlayTraffic(
            br.traffic_api_pb2.PlaybackInfos(playbackInfo=list(map(self.__create_playback_config, playback_list)))
        )
        self.__check_playbackmode_result(status)
        return status

    def record(self, namespace: str, path: str) -> br.traffic_api_pb2.PlaybackInfos:
        playback_list = [
            {
                "namespace": namespace,
                "path": path,
                "mode": br.traffic_api_pb2.Mode.RECORD,
            }
        ]

        status = self.traffic_stub.PlayTraffic(
            br.traffic_api_pb2.PlaybackInfos(playbackInfo=list(map(self.__create_playback_config, playback_list)))
        )
        self.__check_playbackmode_result(status)
        return status

    def stop(self, namespace: str, path: str, silent: bool = False) -> br.traffic_api_pb2.PlaybackInfos:
        playback_list = [
            {
                "namespace": namespace,
                "path": path,
                "mode": br.traffic_api_pb2.Mode.STOP,
            }
        ]

        status: traffic_api.PlaybackInfos = self.traffic_stub.PlayTraffic(
            br.traffic_api_pb2.PlaybackInfos(playbackInfo=list(map(self.__create_playback_config, playback_list)))
        )
        if not silent:
            self.__check_playbackmode_result(status)
        return status

    def listen_on_playback(self, repeat: bool, recording_and_namespace: List[Any], callback: Callable[[int, int, str], None]) -> None:
        # include recording_and_namespace if we want to loop the recording
        # This can probably be improved
        def get_mode(mode: int) -> str:
            if mode == 0:
                return "playing"
            if mode == 1:
                return "paused"
            if mode == 2:
                return "stopped"
            raise ValueError("Unknown Mode")

        sub = self.traffic_stub.PlayTrafficStatus(br.common_pb2.Empty())
        for playback_state in sub:
            # p = typing.cast(br.traffic_api_pb2.PlaybackInfos, playback_state) # REDUNDANT CAST
            p = playback_state
            offset_length = int(p.playbackInfo[0].playbackMode.offsetTime / 1000000)
            start_time = p.playbackInfo[0].playbackMode.startTime
            end_time = p.playbackInfo[0].playbackMode.endTime
            mode = p.playbackInfo[0].playbackMode.mode

            total_length = int((end_time - start_time) / 1000000)

            if mode == 2 and repeat:
                # If we get a stop and is fairly (this is mostly not 100%) close to the end
                # we repeat the recording when files are included
                if abs(total_length - offset_length) < 5:
                    self.play(recording_and_namespace)
            callback(offset_length, total_length, get_mode(mode))

    def listen_on_frame_distribution(self, namespace: str, callback: Callable[[Dict[str, Any]], None]) -> None:
        config = network_api.FramesDistributionConfig(namespace=br.common_pb2.NameSpace(name=namespace))
        frame_distribution_stream = self.network_stub.SubscribeToFramesDistribution(config)
        for frame in frame_distribution_stream:
            f = MessageToDict(frame, including_default_value_fields=True, preserving_proto_field_name=True)
            callback(f)

    def pause(self, namespace: str, path: str, silent: bool = False) -> br.traffic_api_pb2.PlaybackInfos:
        playback_list = [
            {
                "namespace": namespace,
                "path": path,
                "mode": br.traffic_api_pb2.Mode.PAUSE,
            }
        ]

        status: traffic_api.PlaybackInfos = self.traffic_stub.PlayTraffic(
            br.traffic_api_pb2.PlaybackInfos(playbackInfo=list(map(self.__create_playback_config, playback_list)))
        )
        if not silent:
            self.__check_playbackmode_result(status)
        return status

    def stop_multiple(self, namespaces: List[str], path: str) -> br.traffic_api_pb2.PlaybackInfos:
        def to_playback(namespace: str) -> Dict[str, Any]:
            return {
                "namespace": namespace,
                "path": path + "_" + namespace,
                "mode": br.traffic_api_pb2.Mode.STOP,
            }

        playback_list = list(map(to_playback, namespaces))

        status = self.traffic_stub.PlayTraffic(
            br.traffic_api_pb2.PlaybackInfos(playbackInfo=list(map(self.__create_playback_config, playback_list)))
        )
        self.__check_playbackmode_result(status)
        return status

    def diagnose_stop(self, namespace: List[str]) -> None:
        recording_name = "diagnose__"
        self.stop_multiple(namespace, recording_name)

    def diagnose(self, namespace: List[str], wait_for_traffic: bool = False) -> None:
        recording_name = "diagnose__"

        keep_running = True
        keep_running_during_recording = True

        def exit_on_ctrlc(_sig: Any, _frame: Any) -> None:
            nonlocal keep_running
            keep_running = False
            nonlocal keep_running_during_recording
            keep_running_during_recording = False
            # progress.add_task(description=f"Cleaning up, please wait...", total=None)

        os_signal.signal(os_signal.SIGINT, exit_on_ctrlc)

        while keep_running:
            keep_running = wait_for_traffic
            self.record_multiple(namespace, recording_name)
            for i in range(5):
                if keep_running_during_recording:
                    time.sleep(1)

            self.stop_multiple(namespace, recording_name)

            response = []
            with tempfile.TemporaryDirectory() as tmpdirname:
                for ns in namespace:
                    path = recording_name + "_" + ns
                    tmp_file = os.path.join(tmpdirname, path)
                    self.download(path, tmp_file)
                    self.delete_files([path], False)
                    with zipfile.ZipFile(tmp_file, "r") as zip_ref:
                        zip_ref.extractall(tmpdirname)

                        file_stat = os.stat(os.path.join(tmpdirname, path + ".raw"))
                        response.append({"namespace": ns, "data": file_stat.st_size > 0})

            for r in response:
                if r["data"]:
                    print(f"Successfully received traffic on {r['namespace']}")
                    keep_running = False
                elif not wait_for_traffic or (not keep_running and not keep_running_during_recording):
                    print(f"Namespace {r['namespace']} did not receive any traffic")

    def upload(self, file: str, dest: str) -> None:
        try:
            br.helper.upload_file(system_stub=self.system_stub, path=file, dest_path=dest)
        except grpc.RpcError as rpc_error:
            print(f"Failed to upload file - {rpc_error.details()} ({rpc_error.code()})")
            raise typer.Exit(1)

    def delete_files(self, path: List[str], exit_on_faliure: bool) -> None:
        for file in path:
            try:
                self.system_stub.BatchDeleteFiles(
                    br.system_api_pb2.FileDescriptions(
                        fileDescriptions=[br.system_api_pb2.FileDescription(path=file.replace(ntpath.sep, posixpath.sep))]
                    )
                )
            except grpc.RpcError as rpc_error:
                print(f"Failed to delete file - {rpc_error.details()} ({rpc_error.code()})")
                if exit_on_faliure:
                    raise typer.Exit(1)

    # rpc BatchDeleteFiles (FileDescriptions) returns (Empty) {}

    def upload_folder(self, folder: str) -> None:
        try:
            br.helper.upload_folder(system_stub=self.system_stub, folder=folder)
        except grpc.RpcError as rpc_error:
            print(f"Failed to upload file - {rpc_error.details()} ({rpc_error.code()})")
            raise typer.Exit(1)

    def download(self, file: str, dest: str, delegate_err: bool = False) -> None:
        try:
            br_helper.download_file(system_stub=self.system_stub, path=file, dest_path=dest)
        except grpc.RpcError as rpc_error:
            if delegate_err:
                raise rpc_error
            print(f"Failed to download file - {rpc_error.details()} ({rpc_error.code()})")
            # There will be an empty file if the download fails so remove that one here
            os.remove(dest)
            raise typer.Exit(1)

    def reload_config(self) -> None:
        try:
            request = br.common_pb2.Empty()
            response = self.system_stub.ReloadConfiguration(request, timeout=60000)
            if response.errorMessage:
                print(f"Failed to reload config: {response.errorMessage}")
                raise typer.Exit(1)
            # br.helper.reload_configuration(system_stub=self.system_stub)
        except grpc.RpcError as rpc_error:
            print(f"Failed to reload configuration - {rpc_error.details()} ({rpc_error.code()})")
            raise typer.Exit(1)

    def list_namespaces(self) -> List[str]:
        # Lists available signals
        configuration = self.system_stub.GetConfiguration(br.common_pb2.Empty())
        namespaces = []
        for network_info in configuration.networkInfo:
            namespaces.append(network_info.namespace.name)
        return namespaces

    def list_signal_names(self, prefix: Union[str, None], suffix: Union[str, None]) -> List[Dict[str, Any]]:
        # Lists available signals
        configuration = self.system_stub.GetConfiguration(br.common_pb2.Empty())

        signal_names = []
        for network_info in configuration.networkInfo:
            res = self.system_stub.ListSignals(network_info.namespace)
            for finfo in res.frame:
                if (prefix is None or finfo.signalInfo.id.name.startswith(prefix)) and (
                    suffix is None or finfo.signalInfo.id.name.endswith(suffix)
                ):
                    metadata_dict = MessageToDict(
                        finfo.signalInfo.metaData,
                        including_default_value_fields=True,
                        preserving_proto_field_name=True,
                    )
                    sig_dict = {
                        "signal": finfo.signalInfo.id.name,
                        "namespace": network_info.namespace.name,
                    }
                    signal_names.append({**sig_dict, **metadata_dict})

                for sinfo in finfo.childInfo:
                    # For signals we can simply skip if prefix and suffix exists does not match
                    if (prefix is not None and not sinfo.id.name.startswith(prefix)) or (
                        suffix is not None and not sinfo.id.name.endswith(suffix)
                    ):
                        continue

                    metadata_dict = MessageToDict(
                        sinfo.metaData,
                        including_default_value_fields=True,
                        preserving_proto_field_name=True,
                    )
                    sig_dict = {
                        "signal": sinfo.id.name,
                        "namespace": network_info.namespace.name,
                    }
                    signal_names.append({**sig_dict, **metadata_dict})

        return signal_names

    def subscribe_on_script(
        self,
        script: bytes,
        on_frame: Callable[[Sequence[br.network_api_pb2.Signal]], None],
        changed_values_only: bool = False,
    ) -> Any:
        client_id = br.common_pb2.ClientId(id="cli")
        # sync = queue.Queue()
        thread = Thread(
            target=br.act_on_scripted_signal,
            args=(
                client_id,
                self.network_stub,
                script,
                changed_values_only,  # True: only report when signal changes
                lambda frame: self.__each_signal(frame, on_frame),
                lambda sub: (self.q.put(("cli", sub))),
            ),
        )
        thread.start()
        # wait for subscription to settle
        return self.q.get()

    def validate_and_get_subscribed_signals(
        self, subscribed_namespaces: List[str], subscribed_signals: List[str]
    ) -> List[SubscribableSignal]:
        # Since we cannot know which list[signals] belongs to which namespace we need to fetch
        # all signals from the broker and find the proper signal with namespace. Finally we
        # also filter out namespaces that we do not need since we might have duplicated signal names
        # over namespaces
        # Begin

        def verify_namespace(available_signal: List[Dict[str, str]]) -> List[str]:
            return list(filter(lambda namespace: available_signal["namespace"] == namespace, subscribed_namespaces))  # type: ignore

        def find_subscribed_signal(available_signal: List[Dict[str, str]]) -> List[str]:
            return list(filter(lambda s: available_signal["signal"] == s, subscribed_signals))  # type: ignore

        existing_signals = self.list_signal_names(prefix=None, suffix=None)
        existing_ns = set(map(lambda s: s["namespace"], existing_signals))
        ns_not_matching = []
        for ns in subscribed_namespaces:
            if ns not in existing_ns:
                ns_not_matching.append(ns)
        if len(ns_not_matching) > 0:
            ErrorPrinter.print_hint(f"Namespace(s) {ns_not_matching} does not exist on broker. Namespaces found on broker: {existing_ns}")
            sys.exit(1)

        available_signals = list(filter(verify_namespace, existing_signals))  # type: ignore
        signals_to_subscribe_to = list(filter(find_subscribed_signal, available_signals))  # type: ignore

        # Check if subscription is done on signal that is not in any of these namespaces
        signals_subscribed_to_but_does_not_exist = set(subscribed_signals) - set(map(lambda s: s["signal"], signals_to_subscribe_to))

        if len(signals_subscribed_to_but_does_not_exist) > 0:
            ErrorPrinter.print_hint(f"One or more signals you subscribed to does not exist {signals_subscribed_to_but_does_not_exist}")
            sys.exit(1)

        return list(map(lambda s: SubscribableSignal(s["signal"], s["namespace"]), signals_to_subscribe_to))

    def long_name_subscribe(
        self, signals_to_subscribe_to: List[SubscribableSignal], on_frame: Callable[..., Any], changed_values_only: bool = True
    ) -> Any:
        client_id = br.common_pb2.ClientId(id="cli")

        # TODO - This can be improved moving forward and we also need to move the validation into api
        self.validate_and_get_subscribed_signals(
            list(map(lambda s: s.namespace, signals_to_subscribe_to)), (list(map(lambda s: s.name, signals_to_subscribe_to)))
        )

        def to_protobuf_signal(s: SubscribableSignal) -> br.common_pb2.SignalId:
            return self.signal_creator.signal(s.name, s.namespace)

        signals_to_subscribe_on = list(map(to_protobuf_signal, signals_to_subscribe_to))

        Thread(
            target=br.act_on_signal,
            args=(
                client_id,
                self.network_stub,
                signals_to_subscribe_on,
                changed_values_only,  # True: only report when signal changes
                lambda frame: self.__each_signal(frame, on_frame),
                lambda sub: (self.q.put(("cloud_demo", sub))),
            ),
        ).start()
        # Wait for subscription
        ecu, subscription = self.q.get()
        return subscription

    def subscribe(
        self,
        subscribed_signals: list[str],
        subscribed_namespaces: list[str],
        on_frame: Callable[..., Any],
        changed_values_only: bool = True,
    ) -> Any:
        client_id = br.common_pb2.ClientId(id="cli")

        signals_to_subscribe_to: List[SubscribableSignal] = self.validate_and_get_subscribed_signals(
            subscribed_namespaces, subscribed_signals
        )

        def to_protobuf_signal(s: SubscribableSignal) -> br.common_pb2.SignalId:
            return self.signal_creator.signal(s.name, s.namespace)

        signals_to_subscribe_on = list(map(to_protobuf_signal, signals_to_subscribe_to))

        Thread(
            target=br.act_on_signal,
            args=(
                client_id,
                self.network_stub,
                signals_to_subscribe_on,
                changed_values_only,  # True: only report when signal changes
                lambda frame: self.__each_signal(frame, on_frame),
                lambda sub: (self.q.put(("cloud_demo", sub))),
            ),
        ).start()
        # Wait for subscription
        ecu, subscription = self.q.get()
        return subscription

    def __each_signal(self, signals: Iterable[br.network_api_pb2.Signal], callback: Callable[..., Any]) -> None:
        callback(
            map(
                lambda s: {"timestamp_us": s.timestamp, "namespace": s.id.namespace.name, "name": s.id.name, "value": self.__get_value(s)},
                signals,
            )
        )

    @staticmethod
    def __get_value(signal: br.network_api_pb2.Signal) -> Any:
        if signal.raw != b"":
            return "0x" + binascii.hexlify(signal.raw).decode("ascii")
        if signal.HasField("integer"):
            return signal.integer
        if signal.HasField("double"):
            return signal.double
        if signal.HasField("arbitration"):
            return signal.arbitration
        return "empty"

    @staticmethod
    def __create_playback_config(item: Dict[str, Any]) -> br.traffic_api_pb2.PlaybackInfo:
        """Creating configuration for playback

        Parameters
        ----------
        item : dict
            Dictionary containing 'path', 'namespace' and 'mode'

        Returns
        -------
        PlaybackInfo
            Object instance of class

        """

        def get_offset_time() -> int:
            if "offsettime" in item:
                return int(item["offsettime"])
            return 0

        playback_config = br.traffic_api_pb2.PlaybackConfig(
            fileDescription=br.system_api_pb2.FileDescription(path=item["path"]),
            namespace=br.common_pb2.NameSpace(name=item["namespace"]),
        )
        return br.traffic_api_pb2.PlaybackInfo(
            playbackConfig=playback_config,
            playbackMode=br.traffic_api_pb2.PlaybackMode(mode=item["mode"], offsetTime=get_offset_time()),
        )

    def get_license(self) -> LicenseInfo:
        license_info = self.system_stub.GetLicenseInfo(br.common_pb2.Empty())
        return LicenseInfo(
            valid=license_info.status == br.system_api_pb2.LicenseStatus.VALID,
            expires=license_info.expires,
            email=license_info.requestId,
            machine_id=license_info.requestMachineId.decode("utf-8"),
        )

    def apply_license(self, license_data_b64: bytes) -> LicenseInfo:
        license = br.system_api_pb2.License()
        license.data = license_data_b64
        license.termsAgreement = True
        self.system_stub.SetLicense(license)
        return self.get_license()



from __future__ import annotations

import os
from time import sleep

import typer
from zeroconf import (
    IPVersion,
    ServiceBrowser,
    ServiceStateChange,
    Zeroconf,
)

from cli.typer import typer_utils

from . import export, files, licenses, playback, record, scripting, signals

app = typer_utils.create_typer(rich_markup_mode="rich")


@app.callback()
def main(
    url: str = typer.Option(None, is_eager=False, help="Broker URL", envvar="REMOTIVE_BROKER_URL"),
) -> None:
    # This can be used to override the --url per command, lets see if this is a better approach
    if url is not None:
        os.environ["REMOTIVE_BROKER_URL"] = url
    # Do other global stuff, handle other global options here


@app.command(help="Discover brokers on this network")
def discover() -> None:
    # print("Not implemented")

    zeroconf = Zeroconf(ip_version=IPVersion.V4Only)

    services = ["_remotivebroker._tcp.local."]
    # services = list(ZeroconfServiceTypes.find(zc=zeroconf))

    print("\nLooking for RemotiveBrokers on your network, press Ctrl-C to exit...\n")
    ServiceBrowser(zeroconf, services, handlers=[on_service_state_change])

    try:
        while True:
            sleep(0.1)
    except KeyboardInterrupt:
        pass
    finally:
        zeroconf.close()


def on_service_state_change(zeroconf: Zeroconf, service_type: str, name: str, state_change: ServiceStateChange) -> None:
    # print(f"Service {name} state changed: {state_change}")

    if state_change is ServiceStateChange.Removed:
        print(f"Service {name} was removed")

    if state_change is ServiceStateChange.Updated:
        print(f"Service {name} was updated")

    if state_change is ServiceStateChange.Added:
        print(f"[ {name} ]")
        info = zeroconf.get_service_info(service_type, name)
        # print("Info from zeroconf.get_service_info: %r" % (info))

        if info:
            # addresses = ["%s:%d" % (addr, cast(int, info.port)) for addr in info.parsed_scoped_addresses()]
            for addr in info.parsed_scoped_addresses():
                print(f"RemotiveBrokerApp: http://{addr}:8080")
                print(f"RemotiveBroker http://{addr}:50051")
            # print("  Weight: %d, priority: %d" % (info.weight, info.priority))
            # print(f"  Server: {info.server}")
            # if info.properties:
            #    print("  Properties are:")
            #    for key, value in info.properties.items():
            #        print(f"    {key}: {value}")
            # else:
            #    print("  No properties")
        else:
            print("  No info")
        print("\n")


app.add_typer(playback.app, name="playback", help="Manage playing recordings")
app.add_typer(record.app, name="record", help="Record data on buses")
app.add_typer(files.app, name="files", help="Upload/Download configurations and recordings")
app.add_typer(signals.app, name="signals", help="Find and subscribe to signals")
app.add_typer(export.app, name="export", help="Export to external formats")
app.add_typer(scripting.app, name="scripting", help="LUA scripting utilities")
app.add_typer(licenses.app, name="license", help="View and request license to broker")

if __name__ == "__main__":
    app()



import dataclasses
import json
from enum import Enum

import typer
from rich import print_json

from cli.typer import typer_utils

from .license_flows import LicenseFlow

app = typer_utils.create_typer()

help_text = """
    More info on our docs page
    https://docs.remotivelabs.com/docs/remotive-broker/getting-started

    --connect describes how the connection is made to the broker and helps us properly connect to broker

      url      = Connects to the broker with the specified url, this is
      hotspot  = If you are using RemotiveBox (our reference hardware) you can connect to the
                 broker over its wifi hotspot. Getting the license from a RemotiveBox over its wifi hotspot
                 requires you switch wi-fi network to the RemotiveBox hotspot called 'remotivelabs-xxx' where
                 'xxx' is a random generated id.

    --url is the broker url, this is mandatory when connect type is "url"
    """


class Connection(str, Enum):
    hotspot = "hotspot"
    url = "url"
    # TODO - Support discover using mdns


@app.command()
def describe(
    connect: Connection = typer.Option("url", case_sensitive=False, help="How to connect to broker"),
    url: str = typer.Option("http://localhost:50051", is_eager=False, help="Broker URL", envvar="REMOTIVE_BROKER_URL"),
) -> None:
    """
    Show licence information

    More info on our docs page
    https://docs.remotivelabs.com/docs/remotive-broker/getting-started

    --connect describes how the connection is made to the broker and helps us properly connect to broker

    url      = Connects to the broker with the specified url, this is
    hotspot  = If you are using RemotiveBox (our reference hardware) you can connect to the
               broker over its wifi hotspot. Getting the license from a RemotiveBox over its wifi hotspot
               requires you switch wi-fi network to the RemotiveBox hotspot called 'remotivelabs-xxx' where
               'xxx' is a random generated id.

    --url is the broker url, this is mandatory when connect type is "url"
    """
    license_flow = LicenseFlow()
    if connect == Connection.url:
        existing_license = license_flow.describe_with_url(url)
        print_json(json.dumps(dataclasses.asdict(existing_license)))
    if connect == Connection.hotspot:
        existing_license = license_flow.describe_with_hotspot(url if url != "http://localhost:50051" else None)
        print_json(json.dumps(dataclasses.asdict(existing_license)))


@app.command()
def request(
    connect: Connection = typer.Option("url", case_sensitive=False, help="How to connect to broker"),
    url: str = typer.Option(
        "http://localhost:50051",
        is_eager=False,
        help="Broker url, this is mandatory when connect type is 'url'",
        envvar="REMOTIVE_BROKER_URL",
    ),
) -> None:
    """
    Requests and applies a new or existing License to a broker, Note that internet access is required on your
    computer

    More info on our docs page
    https://docs.remotivelabs.com/docs/remotive-broker/getting-started

    --connect describes how the connection is made to the broker and helps us properly connect to broker

    url      = Use url to connect to broker (use --url)
    hotspot  = If you are using RemotiveBox (our reference hardware) you can connect to the
               broker over its wifi hotspot. Licensing a broker on a RemotiveBox over its wifi hotspot
               requires you switch wi-fi network to the RemotiveBox hotspot called 'remotivelabs-xxx' where
               'xxx' is a random generated id.
               https://docs.remotivelabs.com/docs/remotive-broker/getting-started/remotive-box

    --url is the broker url, this is mandatory when connect type is "url"
    """

    license_flow = LicenseFlow()
    if connect == Connection.url:
        license_flow.request_with_url_with_internet(url)
    if connect == Connection.hotspot:
        license_flow.request_with_hotspot(url if url != "http://localhost:50051" else None)



from __future__ import annotations

import dataclasses
import json
from dataclasses import dataclass
from datetime import date, datetime
from typing import Any, Literal

DEFAULT_EMAIL = "unknown@remotivecloud.com"

TokenType = Literal["authorized_user", "service_account"]


def _parse_date(date_str: str) -> date:
    normalized = date_str.replace("Z", "+00:00")
    return datetime.fromisoformat(normalized).date()


def _parse_token_type(token: str) -> TokenType:
    if token.startswith("pa"):
        return "authorized_user"
    if token.startswith("sa"):
        return "service_account"
    raise ValueError(f"Unknown token type for token: {token}")


def _from_dict(d: dict[str, Any]) -> TokenFile:
    if "version" not in d:
        token_type = _parse_token_type(d["token"])
        return TokenFile(
            version="1.0",
            type=token_type,
            name=d["name"],
            token=d["token"],
            created=_parse_date(d["created"]),
            expires=_parse_date(d["expires"]),
            account=TokenFileAccount(email=DEFAULT_EMAIL),
        )

    account_email = d.get("account", {}).get("email", DEFAULT_EMAIL)
    return TokenFile(
        version=d["version"],
        type=d["type"],
        name=d["name"],
        token=d["token"],
        created=_parse_date(d["created"]),
        expires=_parse_date(d["expires"]),
        account=TokenFileAccount(email=account_email),
    )


def loads(data: str) -> TokenFile:
    return _from_dict(json.loads(data))


def dumps(token: TokenFile) -> str:
    return json.dumps(dataclasses.asdict(token), default=str)


@dataclass
class TokenFileAccount:
    email: str


@dataclass
class TokenFile:
    version: str
    type: TokenType
    name: str
    token: str
    created: date
    expires: date
    account: TokenFileAccount

    def is_expired(self) -> bool:
        return datetime.today().date() > self.expires

    def expires_in_days(self) -> int:
        return (self.expires - datetime.today().date()).days

    @staticmethod
    def from_json_str(data: str) -> TokenFile:
        return loads(data)

    @staticmethod
    def from_dict(data: dict[str, Any]) -> TokenFile:
        return _from_dict(data)
